from bs4 import BeautifulSoup
import re
from email.utils import parsedate_to_datetime
from datetime import datetime
import time
import pytz


def third(n) -> int:
    """Calculates one-third of a given value.

    :param n the integer to calculate one-third of
    :returns int: one-third of n, rounded down
    """
    return int(n / 3)


def median(arr):
    """Determines the median of a list of numbers.

    :param arr a list of ints and/or floats of which to determine the median
    :returns int or float: the median value in arr
    """
    if len(arr) == 0:
        return None

    arr_sorted = sorted(arr)
    midpoint = int(len(arr) / 2)

    if len(arr) % 2 == 0:
        # even number of elements; get the average of the middle two
        result = (arr_sorted[midpoint - 1] + arr_sorted[midpoint]) / 2
    else:
        result = arr_sorted[midpoint]
    return result


def sanitize_path(path) -> str:
    """Replaces any characters in path that the file system may not support.

    This method replaces any non-alphanumeric characters with an underscore,
    with the exception of hyphens.

    :param path the original path
    :returns str: the given path with potentially unsafe characters replaced
    """
    # adapted from https://stackoverflow.com/a/13593932
    path = re.sub("[^\\w\\-]", "_", path)
    return path


def is_true(string) -> bool:
    """Determines whether a string represents True.

    As the name suggests, any input which is not explicitly evaluated to True
    will cause this method to return False.

    :param string the string to evaluate
    """
    assert isinstance(string, str)

    return string in ["True", "true", "1"]


def html_to_plain(html) -> str:
    """Converts a potentially HTML-formatted string to user-friendly plaintext.

    :param html the text to convert with potential html tags
    :returns str: the given text with html tags removed
    """
    soup = BeautifulSoup(html, "html.parser")
    return soup.get_text()


def datetime_from_rfc822(date) -> datetime:
    """Convert a date string in RFC822 format into a datetime.

    https://www.w3.org/Protocols/rfc822/
    https://validator.w3.org/feed/docs/error/InvalidRFC2822Date.html

    :param date string for the date/time in RFC822 format
    :returns datetime: a matching datetime object, or -1
    """
    try:
        return parsedate_to_datetime(date).replace(tzinfo=pytz.UTC)
    except (TypeError, ValueError):
        return -1


def seconds_to_time(seconds: int) -> str:
    seconds = max(0, seconds)
    return time.strftime("%H:%M:%S", time.gmtime(seconds))



import curses
import sys
import threading
import re
import argparse
import os
import io
import ctypes
import tempfile

from gevent import monkey

monkey.patch_all(thread=False, select=False)

import castero
from castero import helpers
from castero.config import Config
from castero.database import Database
from castero.display import Display
from castero.feed import Feed
from castero.subscriptions import Subscriptions


def import_subscriptions(path: str, database: Database) -> None:
    subscriptions = Subscriptions()

    # Load may raise an error, but they are user-friendly enough that we don't
    # need to catch them here. It's also okay to crash at this point.
    subscriptions.load(path)

    for generated in subscriptions.parse():
        if isinstance(generated, Feed):
            feed = generated
            database.replace_feed(feed)
            episodes = feed.parse_episodes()
            database.replace_episodes(feed, episodes)
            print('Added "%s" with %d episodes' % (str(feed), len(episodes)))
        else:
            print("ERROR: Failed to load %s -- %s" % (str(generated[0]), str(generated[1])))

    database.close()
    print("Imported %d feeds" % len(subscriptions.feeds))


def export_subscriptions(path: str, database: Database) -> None:
    subscriptions = Subscriptions()

    feeds = database.feeds()
    subscriptions.generate(feeds)
    # Save may raise an error, but they are user-friendly enough that we don't
    # need to catch them here. It's also okay to crash at this point.
    subscriptions.save(path)

    print("Exported %d feeds" % len(feeds))


def redirect_stderr() -> io.TextIOWrapper:
    temp_file = tempfile.TemporaryFile(prefix="%s-" % castero.__title__)

    # get os-specific stderr descriptor
    stderr = "stderr"
    if sys.platform == "darwin":
        stderr = "__stderrp"

    libc = ctypes.CDLL(None)
    c_stderr = ctypes.c_void_p.in_dll(libc, stderr)

    stderr_fd = sys.stderr.fileno()
    libc.fflush(c_stderr)
    sys.stderr.close()

    # make the stderr fd point to the temp_file
    os.dup2(temp_file.fileno(), stderr_fd)

    # overwrite sys.stderr to use our modified fd
    # - not explicitly necessary for our purposes, since curses does not
    #   use this field
    sys.stderr = io.TextIOWrapper(os.fdopen(stderr_fd, "wb"))

    return temp_file.fileno()


def main():
    database = Database()

    # parse command line arguments
    parser = argparse.ArgumentParser(prog=castero.__title__, description=castero.__description__)
    parser.add_argument(
        "-V", "--version", action="version", version="%(prog)s {}".format(castero.__version__)
    )
    parser.add_argument("--import", help="path to OPML file of feeds to add")
    parser.add_argument("--export", help="path to save feeds as OPML file")
    args = parser.parse_args()

    if vars(args)["import"] is not None:
        import_subscriptions(vars(args)["import"], database)
        sys.exit(0)
    elif vars(args)["export"] is not None:
        export_subscriptions(vars(args)["export"], database)
        sys.exit(0)

    # update fields in help menu text
    for field in Config:
        if "{%s}" % field in castero.__help__:
            castero.__help__ = castero.__help__.replace("{%s}" % field, Config[field].ljust(11))
        elif "{%s|" % field in castero.__help__:
            field2 = castero.__help__.split("{%s|" % field)[1].split("}")[0]
            castero.__help__ = castero.__help__.replace(
                "{%s|%s}" % (field, field2), ("%s or %s" % (Config[field], Config[field2])).ljust(11)
            )
        elif "{%s/" % field in castero.__help__:
            field2 = castero.__help__.split("{%s/" % field)[1].split("}")[0]
            castero.__help__ = castero.__help__.replace(
                "{%s/%s}" % (field, field2), ("%s/%s" % (Config[field], Config[field2])).ljust(11)
            )
    remaining_brace_fields = re.compile("\\{.*?\\}").findall(castero.__help__)
    for field in remaining_brace_fields:
        adjusted = field.replace("{", "").replace("}", "").ljust(11)
        castero.__help__ = castero.__help__.replace(field, adjusted)

    # instantiate display
    redirect_stderr()
    stdscr = curses.initscr()
    display = Display(stdscr, database)
    display.clear()
    display.update_parent_dimensions()

    # check if we need to start reloading
    if helpers.is_true(Config["reload_on_start"]):
        reload_thread = threading.Thread(target=database.reload, args=[display])
        reload_thread.start()

    # run initial display operations
    display.display_all()
    display._menus_valid = False
    display._update_timer = 0

    # core loop for the client
    running = True
    while running:
        display.display()
        char = display.getch()
        if char != -1:
            running = display.handle_input(char)

    sys.exit(0)


main()



import json
import os
import sys
import sqlite3
import grequests
from typing import List
from io import StringIO

from castero import helpers
from castero.config import Config
from castero.datafile import DataFile
from castero.episode import Episode
from castero.feed import Feed, FeedError
from castero.queue import Queue
from castero.net import Net


class Database:
    """The user's database.

    This class provides an API for storing and retrieving data from an sqlite
    database file.

    The database replaces the Feeds class which stored feed/episode data in a
    JSON file. We automatically migrate users from that format to the database:
    see _create_from_old_feeds().

    For schema details, see $PACKAGE/templates/migrations.
    """

    PATH = os.path.join(DataFile.DATA_DIR, "castero.db")
    OLD_PATH = os.path.join(DataFile.DATA_DIR, "feeds")
    MIGRATIONS_DIR = os.path.join(DataFile.PACKAGE, "templates/migrations")

    SQL_EPISODES_BY_FEED_WITH_PROGRESS = "select episode.id, episode.title, episode.description, episode.link, episode.pubdate, episode.copyright, episode.enclosure, episode.played, progress.time from episode left join progress on episode.id=progress.ep_id where feed_key=? order by episode.id"
    SQL_EPISODES_WITH_PROGRESS = "select episode.feed_key, episode.id, episode.title, episode.description, episode.link, episode.pubdate, episode.copyright, episode.enclosure, episode.played, progress.time from episode left join progress on episode.id=progress.ep_id order by episode.id"
    SQL_EPISODES_BY_ID = "select episode.feed_key, episode.id, episode.title, episode.description, episode.link, episode.pubdate, episode.copyright, episode.enclosure, episode.played, progress.time from episode left join progress on episode.id=progress.ep_id where episode.id=?"
    SQL_UNPLAYED_EPISODES_BY_FEED = "select episode.id, episode.title, episode.description, episode.link, episode.pubdate, episode.copyright, episode.enclosure, episode.played, progress.time from episode left join progress on episode.id=progress.ep_id where feed_key=? and played=0 order by episode.id"
    SQL_EPISODE_REPLACE = "replace into episode (id, title, feed_key, description, link, pubdate, copyright, enclosure, played)\nvalues (?,?,?,?,?,?,?,?,?)"
    SQL_EPISODE_REPLACE_NOID = "replace into episode (title, feed_key, description, link, pubdate, copyright, enclosure, played)\nvalues (?,?,?,?,?,?,?,?)"
    SQL_FEEDS_ALL = (
        "select key, title, description, link, last_build_date, copyright from feed order by lower(title)"
    )
    SQL_FEED_BY_KEY = (
        "select key, title, description, link, last_build_date, copyright from feed where key=?"
    )
    SQL_FEED_REPLACE = (
        "replace into feed (key, title, description, link, last_build_date, copyright)\nvalues (?,?,?,?,?,?)"
    )
    SQL_FEED_DELETE = "delete from feed where key=?"
    SQL_QUEUE_ALL = "select id, ep_id from queue"
    SQL_QUEUE_REPLACE = "replace into queue (id, ep_id)\nvalues (?,?)"
    SQL_QUEUE_DELETE = "delete from queue"
    SQL_EPISODE_PROGRESS_REPLACE = "replace into progress (ep_id, time)\nvalues (?,?)"
    SQL_EPISODE_PROGRESS_DELETE = "delete from progress where ep_id=?"

    def __init__(self):
        """
        If the database file does not exist but the old Feeds file does, we
        create the database using the old format.
        """
        existed = os.path.exists(self.PATH)
        DataFile.ensure_path(self.PATH)

        self._using_memory = not helpers.is_true(Config["restrict_memory_usage"])

        file_conn = sqlite3.connect(self.PATH, check_same_thread=False)

        if self._using_memory:
            memory_conn = sqlite3.connect(":memory:", check_same_thread=False)
            self._copy_database(file_conn, memory_conn)
            self._conn = memory_conn
        else:
            self._conn = file_conn

        if not existed and os.path.exists(self.OLD_PATH):
            self._create_from_old_feeds()

        self._conn.execute("PRAGMA foreign_keys = ON")
        self.migrate()

    def close(self):
        """Close the database.

        If we were using an in-memory copy of the data, it is written
        to the database file here.
        """
        if self._using_memory:
            DataFile.ensure_path(self.PATH)
            os.rename(self.PATH, self.PATH + ".old")

            file_conn = sqlite3.connect(self.PATH)
            self._copy_database(self._conn, file_conn)
        self._conn.close()

    def migrate(self):
        """Apply SQL migrations.

        Migrations are defined in $PACKAGE/templtaes/migrations.
        """
        cursor = self._conn.cursor()
        cur_version = cursor.execute("pragma user_version").fetchone()[0]

        migration_files = list(os.listdir(self.MIGRATIONS_DIR))
        for migration in sorted(migration_files):
            version = int(migration.split("-")[0])
            if version > cur_version:
                path = os.path.join(self.MIGRATIONS_DIR, migration)
                with open(path, "rt") as f:
                    cursor.executescript(f.read())

    def _copy_database(self, from_connection, to_connection):
        """Copy database contents from one connection to another."""
        if sys.version_info.major == 3 and sys.version_info.minor >= 7:
            from_connection.backup(to_connection)
            return

        cursor = from_connection.cursor()
        cur_version = cursor.execute("pragma user_version").fetchone()[0]
        tempfile = StringIO()
        for line in from_connection.iterdump():
            tempfile.write("%s\n" % line)
        from_connection.close()
        tempfile.seek(0)

        to_connection.execute("PRAGMA user_version = " + str(cur_version))
        to_connection.cursor().executescript(tempfile.read())
        to_connection.commit()
        to_connection.execute("PRAGMA foreign_keys = ON")

    def _create_from_old_feeds(self):
        """Create database from deprecated Feeds format.

        This method is only necessary for users updating to version 0.4.0+ from
        an earlier version.
        """
        self.migrate()
        cursor = self._conn.cursor()

        with open(self.OLD_PATH, "r") as f:
            content = json.loads(f.read())

        for key in content:
            feed_dict = content[key]

            cursor.execute(
                self.SQL_FEED_REPLACE,
                (
                    key,
                    feed_dict["title"],
                    feed_dict["description"],
                    feed_dict["link"],
                    feed_dict["last_build_date"],
                    feed_dict["copyright"],
                ),
            )

            for episode_dict in feed_dict["episodes"]:
                cursor.execute(
                    self.SQL_EPISODE_REPLACE_NOID,
                    (
                        episode_dict["title"],
                        key,
                        episode_dict["description"],
                        episode_dict["link"],
                        episode_dict["pubdate"],
                        episode_dict["copyright"],
                        episode_dict["enclosure"],
                        False,
                    ),
                )

        self._conn.commit()

    def delete_feed(self, feed: Feed) -> None:
        """Delete a feed from the database.

        Note: episodes have a cascade delete relation with their feed.

        :param feed the Feed to delete, which is in the database
        """
        cursor = self._conn.cursor()
        cursor.execute(self.SQL_FEED_DELETE, (feed.key,))
        self._conn.commit()

    def replace_feed(self, feed: Feed) -> None:
        """Replace (or insert) a feed in the database.

        This method is used for both updating a feed and for adding a new one.

        :param feed the Feed to replace
        """
        cursor = self._conn.cursor()
        cursor.execute(
            self.SQL_FEED_REPLACE,
            (feed.key, feed.title, feed.description, feed.link, feed.last_build_date, feed.copyright),
        )
        self._conn.commit()

    def replace_episode(self, feed: Feed, episode: Episode) -> None:
        """Replace (or insert) an episode in the database.

        This method is used for both updating an episode and for adding a new
        one.

        Episode instances have an ep_id field which is the episode's unique id
        in the database, if set. If it was not set when this method is called,
        we update it after the episode has been added to the database.

        :param feed the Feed the episode is a part of
        :param episode the Episode to replace
        """
        cursor = self._conn.cursor()
        if episode.ep_id is None:
            cursor.execute(
                self.SQL_EPISODE_REPLACE_NOID,
                (
                    episode.title,
                    feed.key,
                    episode.description,
                    episode.link,
                    episode.pubdate,
                    episode.copyright,
                    episode.enclosure,
                    episode.played,
                ),
            )
            episode.ep_id = cursor.lastrowid
        else:
            cursor.execute(
                self.SQL_EPISODE_REPLACE,
                (
                    episode.ep_id,
                    episode.title,
                    feed.key,
                    episode.description,
                    episode.link,
                    episode.pubdate,
                    episode.copyright,
                    episode.enclosure,
                    episode.played,
                ),
            )
        self._conn.commit()

    def replace_episodes(self, feed: Feed, episodes: List[Episode]) -> None:
        """Replace (or insert) a list of episodes in the database.

        This method is used for both updating episodes and for adding new ones.

        :param feed the Feed all episode are a part of
        :param episodes a list of Episode's to replace
        """
        cursor = self._conn.cursor()

        # there are different sql queries depending on whether the episode
        # has an id, so we separate them into 2 operations
        episodes_without_id = []
        episodes_with_id = []

        for episode in episodes:
            if episode.ep_id is None:
                episodes_without_id.append(episode)
            else:
                episodes_with_id.append(episode)

        if len(episodes_without_id) > 0:
            cursor.executemany(
                self.SQL_EPISODE_REPLACE_NOID,
                (
                    (
                        episode.title,
                        feed.key,
                        episode.description,
                        episode.link,
                        episode.pubdate,
                        episode.copyright,
                        episode.enclosure,
                        episode.played,
                    )
                    for episode in episodes_without_id
                ),
            )
        if len(episodes_with_id) > 0:
            cursor.executemany(
                self.SQL_EPISODE_REPLACE,
                (
                    (
                        episode.ep_id,
                        episode.title,
                        feed.key,
                        episode.description,
                        episode.link,
                        episode.pubdate,
                        episode.copyright,
                        episode.enclosure,
                        episode.played,
                    )
                    for episode in episodes_with_id
                ),
            )
        self._conn.commit()

    def delete_queue(self) -> None:
        """Clear the queue table."""
        cursor = self._conn.cursor()
        cursor.execute(self.SQL_QUEUE_DELETE)
        self._conn.commit()

    def replace_queue(self, queue: Queue) -> None:
        """Replace the queue in the database.

        This method overwrites the existing queue.

        :param queue the Queue to replace from
        """
        cursor = self._conn.cursor()

        self.delete_queue()

        i = 1
        for player in queue:
            episode = player.episode
            if self.episode(episode.ep_id) is not None:
                cursor.execute(self.SQL_QUEUE_REPLACE, (i, player.episode.ep_id))
                i += 1
        self._conn.commit()

    def feeds(self) -> List[Feed]:
        """Retrieve the list of Feeds.

        :returns List[Feed]: all Feed's in the database
        """
        cursor = self._conn.cursor()
        cursor.execute(self.SQL_FEEDS_ALL)

        feeds = []
        for row in cursor.fetchall():
            feed = Feed(
                url=row[0] if row[0].startswith("http") else None,
                file=row[0] if not row[0].startswith("http") else None,
                title=row[1],
                description=row[2],
                link=row[3],
                last_build_date=row[4],
                copyright=row[5],
            )

            if feed.title:
                feeds.append(feed)
            else:
                self.delete_feed(feed)
        return feeds

    def episodes(self, feed: Feed = None) -> List[Episode]:
        """Retrieve all episodes for a feed.

        :param feed the Feed to retrieve episodes of
        :returns List[Episode]: all Episode's of the given Feed in the database
        """
        cursor = self._conn.cursor()

        if feed is None:
            cursor.execute(self.SQL_EPISODES_WITH_PROGRESS, ())
            rows = cursor.fetchall()

            feed_entries = {}
            for row in rows:
                feed_key = row[0]
                if feed_key not in feed_entries:
                    feed_entries[feed_key] = self.feed(feed_key)

            return [
                Episode(
                    feed_entries[row[0]],
                    ep_id=row[1],
                    title=row[2],
                    description=row[3],
                    link=row[4],
                    pubdate=row[5],
                    copyright=row[6],
                    enclosure=row[7],
                    played=row[8],
                    progress=row[9],
                )
                for row in rows
            ]
        else:
            cursor.execute(self.SQL_EPISODES_BY_FEED_WITH_PROGRESS, (feed.key,))
            rows = cursor.fetchall()
            return self._create_feed_episode_list(feed, rows)

    def unplayed_episodes(self, feed: Feed) -> List[Episode]:
        """Retrieve all unplayed episodes for a feed.

        :param feed the Feed to retrieve episodes of
        :returns List[Episode]: all Episode's of the given Feed in the database
        """
        cursor = self._conn.cursor()
        cursor.execute(self.SQL_UNPLAYED_EPISODES_BY_FEED, (feed.key,))
        rows = cursor.fetchall()
        return self._create_feed_episode_list(feed, rows)

    def _create_feed_episode_list(self, feed: Feed, episode_rows) -> List[Episode]:
        """Create a list of episode from feed episode query rows
        :param feed The feed the episodes are from,
        :param episode_rows Query result rows
        :returns List[Episode]: List of the episodes
        """
        return [
            Episode(
                feed,
                ep_id=row[0],
                title=row[1],
                description=row[2],
                link=row[3],
                pubdate=row[4],
                copyright=row[5],
                enclosure=row[6],
                played=row[7],
                progress=row[8],
            )
            for row in episode_rows
        ]

    def feed(self, key) -> Feed:
        """Retrieve a feed by key.

        :param key the key of the Feed to retrieve, which is the feed's primary
          key in the database
        :returns Feed: the matching Feed, if it exists, or None
        """
        cursor = self._conn.cursor()
        cursor.execute(self.SQL_FEED_BY_KEY, (key,))

        result = cursor.fetchone()
        if result is None:
            return None
        else:
            return Feed(
                url=result[0] if result[0].startswith("http") else None,
                file=result[0] if not result[0].startswith("http") else None,
                title=result[1],
                description=result[2],
                link=result[3],
                last_build_date=result[4],
                copyright=result[5],
                episodes=[],
            )

    def episode(self, ep_id: int) -> Episode:
        """Retrieve an episode by ep_id.

        :param ep_id: the id of the Episode to retrieve, which is the episode's
          primary key in the database
        :returns Episode: the matching Episode, if it exists, or None
        """
        cursor = self._conn.cursor()
        cursor.execute(self.SQL_EPISODES_BY_ID, (ep_id,))

        result = cursor.fetchone()
        if result is None:
            return None
        else:
            feed = self.feed(result[0])
            return Episode(
                feed,
                ep_id=result[1],
                title=result[2],
                description=result[3],
                link=result[4],
                pubdate=result[5],
                copyright=result[6],
                enclosure=result[7],
                played=result[8],
                progress=result[9],
            )

    def queue(self) -> List[Episode]:
        """Retrieve all episodes in the queue.

        :returns List[Episode]: all Episode's in the queue
        """
        cursor = self._conn.cursor()

        # get ep_id's of episodes to retrieve
        cursor.execute(self.SQL_QUEUE_ALL, ())
        ep_ids = [row[1] for row in cursor.fetchall()]

        if len(ep_ids) == 0:
            return []

        # bulk retrieve all episodes
        sql = self.SQL_EPISODES_BY_ID
        for ep_id in ep_ids[1:]:
            sql += " OR id=?"
        cursor.execute(sql, tuple(ep_ids))

        episodes_cache = {}
        feeds_cache = {}
        for result in cursor.fetchall():
            ep_id = result[1]
            if ep_id not in episodes_cache:
                feed_key = result[0]
                feed = feeds_cache[feed_key] if feed_key in feeds_cache else self.feed(feed_key)
                episodes_cache[ep_id] = Episode(
                    feed,
                    ep_id=ep_id,
                    title=result[2],
                    description=result[3],
                    link=result[4],
                    pubdate=result[5],
                    copyright=result[6],
                    enclosure=result[7],
                    played=result[8],
                    progress=result[9],
                )

        # queue may contain repeated ep_id's, so we need to go back to the
        # retrieved ep_ids to get complete list of desired episodes
        return [episodes_cache[ep_id] for ep_id in ep_ids]

    def reload(self, display=None, feeds=None) -> None:
        """Reload feeds in the database.

        To preserve user metadata for episodes (such as played/marked status),
        we use Episode.replace_from() which "manually" copies such fields to
        the new downloaded episode. This is necessary because downloaded
        episodes are new Episode instances and we can't guarantee they have any
        of the same properties.

        Therefore, Episode.replace_from() _must_ be updated if any new user
        metadata fields are added.

        Also: to determine which episodes are the same in order to copy user
        metadata, we simply check whether the string representation of the two
        episodes are matching (usually the episodes' titles). This could cause
        issues if a feed has multiple episodes with the same title, although it
        does not require episode titles to be globally unique (that is,
        episodes with the same name in different feeds will never have issues).

        This method adheres to the max_episodes config parameter to limit the
        number of episodes saved per feed.

        :param display (optional) the display to write status updates to
        :param feeds (optional) a list of feeds to reload. If not specified,
          all feeds in the database will be reloaded
        """
        if feeds is None:
            feeds = self.feeds()
        total_feeds = len(feeds)
        completed_feeds = 0
        errors = 0

        reqs = []
        url_pairs = {}
        file_feeds = []
        # Create async requests for each URL feed. We also keep a map from
        # each feed's URL to the Feed object itself in order to access the
        # object when a request completes (since the response object is all
        # that we are given).
        # We also keep track of file-based feeds, which are handled afterwards.
        for feed in feeds:
            if feed.key.startswith("http"):
                url_pairs[feed.key] = feed
                reqs.append(Net.GGet(feed.key))
            else:
                file_feeds.append(feed)

        # handle each response as downloads complete asynchronously
        for response in grequests.imap(reqs, size=3):
            if display is not None:
                error_str = "(%s errors)" % errors if errors > 0 else ""
                display.change_status(
                    "Reloading feeds (%d/%d) %s" % (completed_feeds, total_feeds, error_str)
                )

            old_feed = None
            response_url = response.request.url
            if response_url in url_pairs:
                old_feed = url_pairs[response.request.url]
            elif hasattr(response, "history") and len(response.history) > 0:
                response_url = response.history[0].url
                old_feed = url_pairs[response_url]
            else:
                errors += 1
                continue

            try:
                new_feed = Feed(url=response_url, response=response)
                self._reload_feed_data(old_feed, new_feed)
                completed_feeds += 1
            except FeedError:
                errors += 1

        # handle each file-based feed
        for old_feed in file_feeds:
            if display is not None:
                error_str = "(%s errors)" % errors if errors > 0 else ""
                display.change_status(
                    "Reloading feeds (%d/%d) %s" % (completed_feeds, total_feeds, error_str)
                )

            try:
                new_feed = Feed(file=old_feed.key)
                self._reload_feed_data(old_feed, new_feed)
                completed_feeds += 1
            except FeedError:
                errors += 1

        if display is not None:
            display.change_status("Successfully reloaded %d feeds" % total_feeds)
            display.menus_valid = False

    def replace_progress(self, episode: Episode, progress: int):
        cursor = self._conn.cursor()
        cursor.execute(self.SQL_EPISODE_PROGRESS_REPLACE, (episode.ep_id, progress))
        episode.progress = progress
        self._conn.commit()

    def delete_progress(self, episode: Episode):
        cursor = self._conn.cursor()
        cursor.execute(self.SQL_EPISODE_PROGRESS_DELETE, (episode.ep_id,))
        episode.progress = None
        self._conn.commit()

    def _reload_feed_data(self, old_feed: Feed, new_feed: Feed):
        """Helper method to update a feed and its episodes in the database.

        :param old_feed the original Feed to be replaced
        :param new_feed a Feed with new/updated data
        """
        # keep user metadata for episodes intact
        new_episodes = new_feed.parse_episodes()
        old_episodes = self.episodes(new_feed)
        episode_progresses = {}
        for new_ep in new_episodes:
            matching_olds = [old_ep for old_ep in old_episodes if str(old_ep) == str(new_ep)]
            if len(matching_olds) == 1:
                new_ep.replace_from(matching_olds[0])
                if matching_olds[0].progress != 0:
                    episode_progresses[str(new_ep)] = new_ep.progress

        # limit number of episodes, if necessary
        max_episodes = int(Config["max_episodes"])
        if max_episodes != -1:
            new_episodes = new_episodes[:max_episodes]

        # update the feed and its episodes in the database
        self.replace_feed(new_feed)
        self.replace_episodes(new_feed, new_episodes)

        # ensure episodes have their progress carried over, if necessary
        added_episodes = self.episodes(new_feed)
        for episode in added_episodes:
            if str(episode) in episode_progresses:
                self.replace_progress(episode, episode.progress)



import curses
import threading

from castero.episode import Episode
from castero.feed import Feed
from castero.menu import Menu
from castero import helpers


class ChronoMenu(Menu):
    """The menu for all episodes in chronological order."""

    def __init__(self, window, source, child=None, active=False) -> None:
        super().__init__(window, source, child=child, active=active)

        self._episodes = []

    def __len__(self) -> int:
        return len(self._filtered_episodes)

    @property
    def _items(self):
        """A list of items in the menu represented as dictionaries."""
        result = []
        for episode in self._filtered_episodes:
            tags = []
            if episode.downloaded:
                tags.append("D")
            if episode.progress > 0:
                tags.append(Episode.PROGRESS_INDICATOR)

            result.append(
                {
                    "attr": curses.color_pair(5) if episode.played else curses.A_NORMAL,
                    "tags": tags,
                    "text": "[%s] %s" % (episode.feed_str, str(episode)),
                }
            )
        return result

    @property
    def title(self) -> str:
        """The title of the menu to display in the window header."""
        base = "Episodes"
        if len(self._filtered_episodes) > 0:
            unplayed_episodes = 0
            for episode in self._filtered_episodes:
                if not episode.played:
                    unplayed_episodes += 1

            return "%s (%d/%d)" % (base, unplayed_episodes, len(self._filtered_episodes))
        return base

    @property
    def item(self) -> Episode:
        """The selected episode."""
        if len(self._filtered_episodes) == 0:
            return None

        return self._filtered_episodes[self._selected]

    @property
    def metadata(self) -> str:
        """Metadata for the selected episode."""
        if len(self._filtered_episodes) == 0:
            return ""

        return self._filtered_episodes[self._selected].metadata

    def update_items(self, obj):
        """Called by the parent menu(the feeds menu) to update our items."""
        super().update_items(obj)

        t = threading.Thread(target=self._request_source_episodes, name="episodes_all")
        t.start()

        self._sanitize()

    def _request_source_episodes(self):
        episodes = self._source.episodes()

        self._episodes = sorted(
            episodes, reverse=not self._inverted, key=lambda ep: helpers.datetime_from_rfc822(ep.pubdate)
        )

        self._sanitize()
        self.display()

    def update_child(self):
        """Not necessary for this menu - - does nothing."""
        pass

    def invert(self):
        """Invert the menu order."""
        super().invert()

        self.update_items(None)

    @property
    def _filtered_episodes(self):
        """A list of episodes which match the menu filter."""
        return list(filter(lambda ep: self._filter_text in str(ep).lower(), self._episodes))



import castero
import requests
import grequests


class Net:
    """Manager for network requests.

    This class provides helper methods for network requests. Generally just a
    wrapper around the requests library.
    """

    USER_AGENT = "%s %s <%s>" % (castero.__title__, castero.__version__, castero.__url__)
    HEADERS = {"User-Agent": USER_AGENT}

    @staticmethod
    def Get(*args, **kwargs) -> requests.models.Response:
        """Send a GET request.

        :param *args arguments for requests.get(); particularly the URL
        :param **kwargs optional arguments for requests.get()
        :returns requests.models.Response: response
        """
        return requests.get(
            *args,
            headers=Net.HEADERS,
            timeout=float(castero.config.Config["request_timeout"]),
            proxies={
                "http": castero.config.Config["proxy_http"],
                "https": castero.config.Config["proxy_https"],
            },
            **kwargs
        )

    @staticmethod
    def GGet(*args, **kwargs):
        return grequests.get(
            *args,
            headers=Net.HEADERS,
            timeout=float(castero.config.Config["request_timeout"]),
            proxies={
                "http": castero.config.Config["proxy_http"],
                "https": castero.config.Config["proxy_https"],
            },
            **kwargs
        )



import time

from castero import constants
from castero.player import Player, PlayerDependencyError


class VLCPlayer(Player):
    """Interface for the vlc media player."""

    NAME = "vlc"

    def __init__(self, title, path, episode) -> None:
        super().__init__(title, path, episode)

        import vlc

        self.vlc = vlc

    @staticmethod
    def check_dependencies():
        """Checks whether dependencies are met for playing a player."""
        try:
            import vlc

            i = vlc.Instance()
            vlc.libvlc_release(i)
        except (ImportError, NameError, OSError, AttributeError):
            raise PlayerDependencyError(
                "Dependency VLC not found, which is required for playing" " media files"
            )

    def _create_player(self) -> None:
        """Creates the player object while making sure it is a valid file."""
        vlc_instance = self.vlc.Instance("--no-video --quiet")

        self._player = vlc_instance.media_player_new()
        self._media = vlc_instance.media_new(self._path)
        self._media.parse()  # may output some junk into the console
        self._player.set_media(self._media)

        self._duration = self._media.get_duration()

    def play(self) -> None:
        """Plays the media."""
        if self._player is None:
            self._create_player()

        self._player.play()
        self._state = 1

    def stop(self) -> None:
        """Stops the media."""
        if self._player is not None:
            if self._player.get_state() == self.vlc.State.Opening:
                self._player.release()
            else:
                self._player.stop()
                self._state = 0

    def pause(self) -> None:
        """Pauses the media."""
        if self._player is not None:
            if self._player.get_state() != self.vlc.State.Opening:
                self._player.pause()
                self._state = 2

    def seek(self, direction, amount) -> None:
        """Seek forward or backward in the media."""
        assert direction == 1 or direction == -1
        if self._player is not None:
            self._player.set_time(
                self._player.get_time() + (direction * amount * constants.MILLISECONDS_IN_SECOND)
            )

    def play_from(self, seconds) -> None:
        """start media from point."""
        self.play()
        self._player.set_time((int)(seconds * constants.MILLISECONDS_IN_SECOND))

    def change_rate(self, direction, display=None) -> None:
        """Increase or decrease the playback speed."""
        assert direction == 1 or direction == -1
        if self._player is not None:
            new_rate = self._player.get_rate() + 0.1 * direction
            self._player.set_rate(new_rate)
            if display:
                display.change_status("Playback speed set to {:0.2f}".format(new_rate))

    def set_rate(self, rate) -> None:
        """Set the playback speed."""
        if self._player is not None:
            self._player.set_rate(rate)

    def set_volume(self, volume) -> int:
        """Set the player volume."""
        if self._player is not None:
            self._player.audio_set_volume(volume)

    @property
    def duration(self) -> int:
        """int: the duration of the player"""
        result = 0
        if self._media is not None:
            self._duration = self._media.get_duration()
            result = self._duration
        return result

    @property
    def volume(self) -> int:
        """int: the volume of the player"""
        if self._player is not None:
            return self._player.audio_get_volume()

    @property
    def time(self) -> int:
        """int: the current time of the player"""
        if self._player is not None:
            return self._player.get_time()

    @property
    def time_str(self) -> str:
        """str: the formatted time and duration of the player"""
        result = "00:00:00/00:00:00"
        if self._player is not None:
            time_seconds = int(self.time / constants.MILLISECONDS_IN_SECOND)
            length_seconds = int(self.duration / constants.MILLISECONDS_IN_SECOND)
            t = time.strftime("%H:%M:%S", time.gmtime(time_seconds))
            d = time.strftime("%H:%M:%S", time.gmtime(length_seconds))
            result = "%s/%s" % (t, d)
        return result



import os
import threading

from castero import constants
from castero import helpers
from castero.config import Config
from castero.datafile import DataFile


class Episode:
    """A single episode from a podcast feed."""

    PROGRESS_INDICATOR = "*"

    def __init__(
        self,
        feed,
        ep_id=None,
        title=None,
        description=None,
        link=None,
        pubdate=None,
        copyright=None,
        enclosure=None,
        played=False,
        progress=None,
    ) -> None:
        """
        At least one of a title or description must be specified.

        :param feed the feed that this episode is a part of
        :param title (optional) the title of the episode
        :param description (optional) the description of the episode
        :param link (optional) a link to the episode
        :param pubdate (optional) the date the episode was published, as a string
        :param copyright (optional) the copyright notice of the episode
        :param enclosure (optional) a url to a media file
        :param played (optional) whether the episode has been played
        """
        assert title is not None or description is not None

        self._feed = feed
        self._ep_id = ep_id
        self._title = title
        self._description = description
        self._link = link
        self._pubdate = pubdate
        self._copyright = copyright
        self._enclosure = enclosure
        self._played = played
        self._progress = progress
        self._downloaded = None

    def __str__(self) -> str:
        """Represent this object as a single-line string.

        :returns string: this episode's title, if it exists, else its description
        """
        if self._title is not None:
            representation = str(self._title)
        else:
            representation = str(self._description)

        representation = representation.split("\n")[0]

        return representation

    def _feed_directory(self) -> str:
        """Gets the path to the downloaded episode's feed directory.

        This method does not ensure whether the directory exists -- it simply
        acts as a single definition of where it _should_ be.

        :returns str: a path to the feed directory
        """
        feed_dirname = helpers.sanitize_path(str(self._feed))
        if Config is None or Config["custom_download_dir"] == "":
            path = DataFile.DEFAULT_DOWNLOADED_DIR
        else:
            path = os.path.expandvars(os.path.expanduser(Config["custom_download_dir"]))
            if not path.startswith("/"):
                path = "/%s" % path
        return os.path.join(path, feed_dirname)

    def get_playable(self) -> str:
        """Gets a playable path for this episode.

        This method checks whether the episode is available on the disk, giving
        the path to that file if so. Otherwise, simply return the episode's
        enclosure, which is probably a URL.

        :returns str: a path to a playable file for this episode
        """
        playable = self.enclosure

        feed_directory = self._feed_directory()
        if os.path.exists(feed_directory):
            for File in os.listdir(feed_directory):
                if File.startswith(str(self.ep_id) + "-"):
                    playable = os.path.join(feed_directory, File)

        return playable

    def download(self, download_queue, display=None):
        """Downloads this episode to the file system.

        This method currently only supports downloading from an external URL.
        In the future, it may be worthwhile to determine whether the episode's
        source is a local file and simply copy it instead.

        :param download_queue the download_queue overseeing this download
        :param display (optional) the display to write status updates to
        """
        if self._enclosure is None:
            if display is not None:
                display.change_status("Download failed: episode does not have" " a valid media source")
            return

        feed_directory = self._feed_directory()
        filename = "%s-%s%s" % (
            self.ep_id,
            helpers.sanitize_path(str(self)),
            str(os.path.splitext(self._enclosure)[1].split("?")[0]),
        )
        output_path = os.path.join(feed_directory, filename)
        DataFile.ensure_path(output_path)

        if display is not None:
            display.change_status("Starting episode download...")

        t = threading.Thread(
            target=DataFile.download_to_file,
            args=[self._enclosure, output_path, str(self), download_queue, display],
            name="download_%s" % str(self),
        )
        t.start()

    def delete(self, display=None):
        """Deletes the episode file from the file system.

        :param display (optional) the display to write status updates to
        """
        if self.downloaded:
            feed_directory = self._feed_directory()
            if os.path.exists(feed_directory):
                for File in os.listdir(feed_directory):
                    if File.startswith(str(self.ep_id) + "-"):
                        os.remove(os.path.join(feed_directory, File))
                        self._downloaded = False
                        if display is not None:
                            display.change_status("Successfully deleted the downloaded episode")

                # if there are no more files in the feed directory, delete it
                if len(os.listdir(feed_directory)) == 0:
                    os.rmdir(feed_directory)

    def check_downloaded(self) -> bool:
        """Check whether the episode is downloaded.

        This method updates the downloaded property.

        :returns bool: whether or not the episode is downloaded
        """
        self._downloaded = False
        feed_directory = self._feed_directory()

        if os.path.exists(feed_directory):
            for File in os.listdir(feed_directory):
                if File.startswith(str(self.ep_id) + "-"):
                    self._downloaded = True
        return self._downloaded

    def replace_from(self, episode) -> None:
        """Replace metadata from the given episode.

        :param episode the source Episode
        """
        self._ep_id = episode._ep_id
        self._played = episode._played
        self._progress = episode._progress

    @property
    def downloaded(self) -> bool:
        """Determines whether the episode is downloaded.

        This method does not guarantee the episode exists, but it determines
        whether it "probably" does. If the download status has not been checked
        since the client started, we check it here and return the result.
        Some methods also update the download status. However, if a file is
        removed externally while the client is still running, the status may
        not be properly updated.

        :returns bool: whether or not the episode is downloaded
        """
        if self._downloaded is None:
            self.check_downloaded()
        return self._downloaded

    @property
    def ep_id(self) -> int:
        """int: the database id of the episode"""
        return self._ep_id

    @ep_id.setter
    def ep_id(self, ep_id) -> None:
        self._ep_id = ep_id

    @property
    def feed_str(self) -> str:
        """str: the string representation of this episode's feed"""
        return str(self._feed)

    @property
    def title(self) -> str:
        """str: the title of the episode"""
        result = self._title
        if result is None:
            result = "Title not available."
        return result

    @property
    def description(self) -> str:
        """str: the description of the episode"""
        result = self._description
        if result is None:
            result = "Description not available."
        return result

    @property
    def link(self) -> str:
        """str: the link of/for the episode"""
        result = self._link
        if result is None:
            result = "Link not available."
        return result

    @property
    def pubdate(self) -> str:
        """str: the publish date of the episode"""
        result = self._pubdate
        if result is None:
            result = "Publish date not available."
        return result

    @property
    def copyright(self) -> str:
        """str: the copyright of the episode"""
        result = self._copyright
        if result is None:
            result = "No copyright specified."
        return result

    @property
    def enclosure(self) -> str:
        """str: the enclosure of the episode"""
        result = self._enclosure
        if result is None:
            result = "Enclosure not available."
        return result

    @property
    def played(self) -> bool:
        """bool: whether the episode has been played"""
        return self._played

    @played.setter
    def played(self, played) -> None:
        self._played = played

    @property
    def progress(self) -> int:
        """int: progress in milliseconds gathered from database"""
        progress = self._progress
        if progress is None:
            progress = 0
        return progress

    @progress.setter
    def progress(self, progress) -> None:
        self._progress = progress

    @property
    def metadata(self) -> str:
        """str: the user-displayed metadata of the episode"""
        description = (
            helpers.html_to_plain(self.description)
            if helpers.is_true(Config["clean_html_descriptions"])
            else self.description
        )
        description = description.replace("\n", "")
        progress = helpers.seconds_to_time(self.progress / constants.MILLISECONDS_IN_SECOND)
        downloaded = (
            "Episode downloaded and available for offline playback."
            if self.downloaded
            else "Episode not downloaded."
        )
        metadata = (
            "!cb{title}\n"
            "{pubdate}\n\n"
            "{link}\n\n"
            "!cbCopyright:\n"
            "{copyright}\n\n"
            "!cbDownloaded:\n"
            "{downloaded}\n\n"
            "!cbDescription:\n"
            "{description}\n\n"
            "!cbTime Played:\n"
            "{progress}\n".format(
                title=self.title,
                pubdate=self.pubdate,
                link=self.link,
                copyright=self.copyright,
                downloaded=downloaded,
                description=description,
                progress=progress,
            )
        )

        return metadata



import curses
import glob
import importlib
import threading
from typing import List
import subprocess
from os.path import dirname, basename, isfile

import castero
from castero import helpers
from castero.config import Config
from castero.database import Database
from castero.downloadqueue import DownloadQueue
from castero.feed import (
    Feed,
    FeedError,
    FeedLoadError,
    FeedDownloadError,
    FeedParseError,
    FeedStructureError,
)
from castero.episode import Episode
from castero.menu import Menu
from castero.perspective import Perspective
from castero.queue import Queue
from castero.player import Player


class DisplayError(Exception):
    """An ambiguous error while handling the display."""


class DisplaySizeError(DisplayError):
    """The display does not have acceptable dimensions."""


class Display:
    """The user-facing display.

    This class is used to handle all user-interaction. It creates and handles
    all aspects of the application's interface, including windows and menus. It
    retrieves input from the user and performs corresponding actions.
    """

    MIN_WIDTH = 20
    MIN_HEIGHT = 8
    INPUT_TIMEOUT = int(Config["refresh_delay"])
    UPDATE_TIMEOUT = 2000 / INPUT_TIMEOUT  # multiple of INPUT_TIMEOUT
    STATUS_TIMEOUT = 3  # multiple of UPDATE_TIMEOUT
    COLOR_NAMES = {
        "black": curses.COLOR_BLACK,
        "blue": curses.COLOR_BLUE,
        "cyan": curses.COLOR_CYAN,
        "green": curses.COLOR_GREEN,
        "magenta": curses.COLOR_MAGENTA,
        "red": curses.COLOR_RED,
        "white": curses.COLOR_WHITE,
        "yellow": curses.COLOR_YELLOW,
        "transparent": -1,
    }
    KEY_MAPPING = {chr(i): i for i in range(256)}
    KEY_MAPPING.update((name[4:], value) for name, value in vars(curses).items() if name.startswith("KEY_"))
    KEY_MAPPING.update({"ENTER": 10, "SPACE": 32})
    AVAILABLE_PLAYERS = {}

    def __init__(self, stdscr, database) -> None:
        """
        :param stdscr a stdscr from curses.initscr()
        :param databse a connected castero.Database
        """
        self._stdscr = stdscr
        self._database = database
        self._parent_x = -1
        self._parent_y = -1
        self._perspectives = {}
        self._active_perspective = int(Config["default_layout"])
        self._header_window = None
        self._footer_window = None
        self._queue = Queue(self)
        self._download_queue = DownloadQueue(self)
        self._status = ""
        self._header_str = ""
        self._footer_str = ""
        self._status_timer = self.STATUS_TIMEOUT
        self._update_timer = self.UPDATE_TIMEOUT
        self._menus_valid = True
        self._modified_episodes = []

        # basic preliminary operations
        self._stdscr.timeout(self.INPUT_TIMEOUT)
        curses.start_color()
        curses.use_default_colors()
        curses.noecho()
        self._curs_set(0)
        curses.cbreak()
        self._stdscr.keypad(True)

        self.update_parent_dimensions()
        self.create_color_pairs()
        self._load_perspectives()
        self._load_players()
        self._restore_queue()
        self._create_windows()
        self.create_menus()

    def color_number(self, color: str) -> int:
        """Convert color string to valid color number"""
        if color in self.COLOR_NAMES:
            return self.COLOR_NAMES[color]
        if color.isnumeric() and int(color) < curses.COLORS:
            return int(color)
        return -1

    def create_color_pairs(self) -> None:
        """Initializes color pairs used for the display.

        Creates the following color pairs (foreground, background):
            - 1: foreground, background
            - 2: background, foreground
            - 3: background_alt, foreground_alt
            - 4: foreground_alt, background_alt
            - 5: foreground_dim, background
        """

        curses.init_pair(
            1,
            self.color_number(Config["color_foreground"]),
            self.color_number(Config["color_background"]),
        )
        curses.init_pair(
            2,
            self.color_number(Config["color_background"]),
            self.color_number(Config["color_foreground"]),
        )
        curses.init_pair(
            3,
            self.color_number(Config["color_background_alt"]),
            self.color_number(Config["color_foreground_alt"]),
        )
        curses.init_pair(
            4,
            self.color_number(Config["color_foreground_alt"]),
            self.color_number(Config["color_background_alt"]),
        )
        curses.init_pair(
            5,
            self.color_number(Config["color_foreground_dim"]),
            self.color_number(Config["color_background"]),
        )
        curses.init_pair(
            6,
            self.color_number(Config["color_foreground_status"]),
            self.color_number(Config["color_background"]),
        )
        curses.init_pair(
            7,
            self.color_number(Config["color_foreground_heading"]),
            self.color_number(Config["color_background"]),
        )
        curses.init_pair(
            8,
            self.color_number(Config["color_foreground_dividers"]),
            self.color_number(Config["color_background"]),
        )

    def _load_perspectives(self) -> None:
        """Load instances of perspectives from the `perspectives` package."""
        # load a list of modules names by manually detecting .py files
        module_files = glob.glob(dirname(__file__) + "/perspectives/*.py")
        module_names = [basename(f)[:-3] for f in module_files if isfile(f)]

        for name in module_names:
            p_mod = importlib.import_module("castero.perspectives.%s" % name)
            p_cls = getattr(p_mod, dir(p_mod)[[cls.lower() == name for cls in dir(p_mod)].index(True)])
            inst = p_cls(self)
            self._perspectives[inst.ID] = inst

    def _load_players(self) -> None:
        """Load player classes from the `players` package."""
        # load a list of modules names by manually detecting .py files
        module_files = glob.glob(dirname(__file__) + "/players/*.py")
        module_names = [basename(f)[:-3] for f in module_files if isfile(f)]

        for name in module_names:
            p_mod = importlib.import_module("castero.players.%s" % name)
            p_cls = getattr(p_mod, dir(p_mod)[[cls.lower() == name for cls in dir(p_mod)].index(True)])
            self.AVAILABLE_PLAYERS[p_cls.NAME] = p_cls

    def _restore_queue(self) -> None:
        """Recreate players in queue from the database."""
        for episode in self.database.queue():
            player = Player.create_instance(
                self.AVAILABLE_PLAYERS, str(episode), episode.get_playable(), episode
            )
            self.queue.add(player)

    def _create_windows(self) -> None:
        """Creates and sets basic parameters for the windows.

        If the windows already exist when this method is run, this method will
        delete them and create new ones.
        """
        # delete old windows if they exist
        if self._header_window is not None:
            del self._header_window
            self._header_window = None
        if self._footer_window is not None:
            del self._footer_window
            self._footer_window = None

        # create windows
        self._header_window = curses.newwin(2, self._parent_x, 0, 0)
        self._footer_window = curses.newwin(2, self._parent_x, self._parent_y - 2, 0)

        # create windows for all perspectives
        for perspective_id in self._perspectives:
            self._perspectives[perspective_id].create_windows()

    def create_menus(self) -> None:
        """Creates the menus used in each window.

        Windows which have menus should be created prior to running this method
        (using _create_windows).
        """
        for perspective_id in self._perspectives:
            self._perspectives[perspective_id].create_menus()

    def show_help(self) -> None:
        """Show the help screen.

        This method takes over the main loop, displaying the screen until a key
        is pressed. This means that typical loop actions, including checking
        the state of the current player, will not run while this screen is up.
        """
        help_lines = castero.__help__.split("\n")

        # the first and last lines that content can be displayed, excluding
        # the final line on the screen which is reserved for the "press any
        # key to exit" line
        y_bounds = (0, self._parent_y - 2)

        # the current vertical offset set by the user with arrow keys
        cur_offset = 0

        # the max offset, which is the offset where the last line in help_lines
        # is displayed at y_bounds[1]
        max_offset = len(help_lines) - (y_bounds[1] - y_bounds[0]) - 2

        help_window = curses.newwin(self._parent_y, self._parent_x, 0, 0)
        help_window.refresh()

        running = True
        update_text = True
        while running:
            inp = self._stdscr.getch()

            if inp == self.KEY_MAPPING["DOWN"]:
                update_text = True
                cur_offset += 1

            elif inp == self.KEY_MAPPING["UP"]:
                update_text = True
                cur_offset -= 1
            elif inp != -1:
                running = False

            if cur_offset > max_offset:
                cur_offset = max_offset
            if cur_offset < 0:
                cur_offset = 0

            if update_text:
                # draw blank lines to clear the screen
                for i in range(self._parent_y):
                    help_window.addstr(i, 0, "".ljust(self._parent_x - 1))

                # add text from help_lines based on offset
                max_lineno = min(len(help_lines), cur_offset + y_bounds[1] + 1)
                for i in range(cur_offset, max_lineno):
                    help_window.addstr(i - cur_offset, 2, help_lines[i][: self._parent_x - 2])

                bottom_line = (
                    "[%d/%d] Press arrow keys to scroll, or any"
                    " other key to exit this screen."
                    % (
                        min(len(help_lines), cur_offset + y_bounds[1] + 2),
                        len(help_lines),
                    )
                )
                help_window.addstr(y_bounds[1] + 1, 2, bottom_line[: self._parent_x - 3])
                help_window.refresh()
                update_text = False

        self.clear()
        self.display_all()

    def display(self) -> None:
        """Draws all windows and sub-features, including titles and borders."""
        # check if the screen size has changed
        self.update_parent_dimensions()

        # decrement the update timer
        self._update_timer -= 1
        if self._update_timer <= 0:
            self._update_timer = self.UPDATE_TIMEOUT
            self.update()

        # display the header and footer
        width = self._header_window.getmaxyx()[1]
        self._header_window.addstr(0, 0, " " * width)
        self._header_window.addstr(0, 0, self._header_str, curses.color_pair(6) | curses.A_BOLD)
        self._footer_window.addstr(1, 0, self._footer_str[: width - 1], curses.color_pair(6) | curses.A_BOLD)

        # add window borders
        self._header_window.hline(
            1, 0, 0, self._header_window.getmaxyx()[1], curses.ACS_HLINE | curses.color_pair(8)
        )
        self._footer_window.hline(
            0, 0, 0, self._footer_window.getmaxyx()[1], curses.ACS_HLINE | curses.color_pair(8)
        )

        # refresh updated windows
        self._footer_window.refresh()
        self._header_window.refresh()

        # add display for current perspective
        self._perspectives[self._active_perspective].display()

    def display_all(self) -> None:
        """Force all perspectives to completely redraw their content.

        For more details, see the description of the method in Perspective.
        """
        self.refresh()
        self._perspectives[self._active_perspective].display_all()
        self.refresh()

    def _get_active_perspective(self) -> Perspective:
        """Retrieve the active/visible Perspective."""
        return self._perspectives[self._active_perspective]

    def _change_active_perspective(self, perspective_id) -> None:
        """Changes _active_perspective to the given perspective.

        :param perspective_id the ID of the perspective to change to
        """
        assert perspective_id in self._perspectives

        self.clear()
        self._active_perspective = perspective_id
        self._perspectives[perspective_id].made_active()
        self.display_all()

    def _get_input_str(self, prompt) -> str:
        """Prompts the user for input and returns the resulting string.

        This method assumes that all input strings will be obtained in the
        footer window.

        :param prompt a string to inform the user of what they need to enter
        :returns str: the user's input
        """
        assert self._footer_window is not None
        assert isinstance(prompt, str)

        self._curs_set(1)
        self._stdscr.timeout(-1)  # disable timeouts while waiting for entry

        # display input prompt
        self._footer_window.addstr(1, 0, " " * (self._footer_window.getmaxyx()[1] - 1))
        self._footer_window.addstr(1, 0, prompt)

        entry_pad = curses.newpad(1, 999)
        current_x = 0
        scroll_x = 0
        input_char = None
        while input_char not in [curses.KEY_ENTER, 10]:
            if input_char is not None:
                # manually handle backspace
                if input_char in [curses.KEY_BACKSPACE, 127]:
                    if current_x > 0:
                        entry_pad.delch(0, current_x - 1)
                        current_x -= 1
                        if scroll_x > 0:
                            scroll_x -= 1
                else:
                    # scroll the input pad if necessary
                    if current_x + len(prompt) > self._footer_window.getmaxyx()[1] - 1:
                        scroll_x += 1

                    # add the entered character to the pad
                    entry_pad.addch(0, current_x, input_char)
                    current_x += 1

                # display current portion of pad
                entry_pad.refresh(
                    0,
                    scroll_x,
                    self._parent_y - 1,
                    len(prompt),
                    self._parent_y - 1,
                    self._footer_window.getmaxyx()[1] - 1,
                )

            # get the next input character
            input_char = self._footer_window.getch()

        self._stdscr.timeout(self.INPUT_TIMEOUT)
        self._footer_window.clear()
        self._curs_set(0)

        return entry_pad.instr(0, 0, entry_pad.getmaxyx()[1]).decode("utf-8").strip()

    def _get_y_n(self, prompt) -> bool:
        """Prompts the user for a yes or no (y/n) input.

        :param prompt a string to inform the user of what they need to enter
        :returns bool: true if the user presses y, false otherwise
        """
        assert self._footer_window is not None
        assert isinstance(prompt, str)

        curses.echo()
        self._curs_set(1)

        self._footer_window.addstr(1, 0, " " * (self._footer_window.getmaxyx()[1] - 1))
        self._footer_window.addstr(1, 0, prompt)
        char = self._footer_window.getch()

        self._footer_window.clear()
        self._curs_set(0)
        curses.noecho()

        return char == ord("y")

    def _curs_set(self, visibility: int) -> None:
        """Safely set the appearance of the cursor.

        :param visibility 0 (invisible), 1 (normal mode) or 2 (high visibility)
        """
        if hasattr(curses, "curs_set"):
            try:
                curses.curs_set(visibility)
            except curses.error:
                pass

    def handle_input(self, c) -> bool:
        """Performs action corresponding to the user's input.

        :param c the input character
        :returns bool: whether or not the application should continue running
        """
        for perspective_id in self._perspectives:
            if c == self.KEY_MAPPING[str(perspective_id)]:
                self._change_active_perspective(perspective_id)

        return self._get_active_perspective().handle_input(c)

    def add_feed(self) -> None:
        """Prompt the user for a feed and add it, if possible."""
        path = self._get_input_str("Enter the URL or path of the feed: ")
        try:
            # assume urls have http in them
            if "http" in path:
                feed = Feed(url=path)
            else:
                feed = Feed(file=path)
            if feed.validated:
                self.database.replace_feed(feed)
                self.database.replace_episodes(feed, feed.parse_episodes())
            self.menus_valid = False
            self.change_status("Feed '%s' successfully added" % str(feed))
        except FeedError as e:
            if isinstance(e, FeedLoadError):
                self.change_status("FeedLoadError: %s" % str(e))
            elif isinstance(e, FeedDownloadError):
                self.change_status("FeedDownloadError: %s" % str(e))
            elif isinstance(e, FeedParseError):
                self.change_status("FeedParseError: %s" % str(e))
            elif isinstance(e, FeedStructureError):
                self.change_status("FeedStructureError: %s" % str(e))
            else:
                self.change_status("FeedError [ambiguous]: %s" % str(e))

    def delete_feed(self, feed: Feed) -> None:
        """Deletes the given feed from the database.

        If the delete_feed_confirmation config option is true, this method will
        first ask for y/n confirmation before deleting the feed.

        Deleting a feed also deletes all downloaded/saved episodes.

        :param feed the Feed to delete, which can be None
        """
        if feed is not None:
            should_delete = True
            if helpers.is_true(Config["delete_feed_confirmation"]):
                should_delete = self._get_y_n("Are you sure you want to delete this feed? (y/n): ")
            if should_delete:
                self.database.delete_feed(feed)
                self.menus_valid = False
                self.change_status("Feed successfully deleted")

    def reload_feeds(self) -> None:
        """Reloads the users' feeds.

        If the total number of feeds is >= the reload_feeds_threshold config
        option, this method will first ask for y/n confirmation.

        This method starts the reloading in a new un-managed thread.
        """
        should_reload = True
        if len(self.database.feeds()) >= int(Config["reload_feeds_threshold"]):
            should_reload = self._get_y_n("Are you sure you want to reload all of your feeds?" " (y/n): ")
        if should_reload:
            t = threading.Thread(target=self.database.reload, args=[self])
            t.start()

    def reload_selected_feed(self, feed: Feed) -> None:
        """Reloads the selected feed.

        This method starts the reloading in a new un-managed thread.
        """
        t = threading.Thread(target=self.database.reload, args=[self, [feed]])
        t.start()

    def save_episodes(self, feed=None, episode=None) -> None:
        """Save a feed or episode.

        Exactly one of either feed or episode must be given.

        :param feed (optional) a feed to download all episodes of
        :param episode (optional) an episode to download
        """
        assert (feed is None or episode is None) and (feed is not episode)

        if feed is not None:
            num_to_save = 0
            for episode in self.database.episodes(feed):
                if not episode.downloaded:
                    num_to_save += 1

            if num_to_save == 0:
                return

            should_delete = self._get_y_n(
                "Are you sure you want to download %d" " episodes from this feed? (y/n): " % num_to_save
            )
            if should_delete:
                for episode in self.database.episodes(feed):
                    if not episode.downloaded:
                        self._download_queue.add(episode)
        else:
            if not episode.downloaded:
                self._download_queue.add(episode)

    def delete_episodes(self, feed=None, episode=None) -> None:
        """Delete a downloaded episode, or all of those from a feed.

        Exactly one of either feed or episode must be given.

        :param feed (optional) a feed to delete all episodes of
        :param episode (optional) an episode or delete
        """
        assert (feed is None or episode is None) and (feed is not episode)

        if feed is not None:
            num_deleted = 0
            num_to_delete = 0
            for episode in self.database.episodes(feed):
                if episode.downloaded:
                    num_to_delete += 1

            if num_to_delete == 0:
                return

            should_delete = self._get_y_n(
                "Are you sure you want to delete %d downloaded"
                " episodes from this feed? (y/n): " % num_to_delete
            )
            if should_delete:
                for episode in self.database.episodes(feed):
                    if episode.downloaded:
                        episode.delete(self)
                        num_deleted += 1
                self.menus_valid = False
                self.change_status("Successfully deleted %d episodes" % num_deleted)
        else:
            if episode.downloaded:
                should_delete = self._get_y_n(
                    "Are you sure you want to delete the downloaded" " episode? (y/n): "
                )
                if should_delete:
                    episode.delete(self)

    def filter_menu(self, menu: Menu) -> None:
        menu.filter_text = self._get_input_str("Filter: ")
        self.menus_valid = False

    def execute_command(self, episode: Episode) -> None:
        """Execute a system command on an episode's enclosure.

        The command to run is set by the user with the execute_command config
        setting. See the description in the config file above that setting for
        details about how it may be formatted.

        :param episode episode to execute the command on
        """
        command = (
            Config["execute_command"]
            .replace("{file}", episode.enclosure)
            .replace("{title}", episode.title)
            .replace("{description}", episode.description)
            .replace("{link}", episode.link)
            .replace("{pubdate}", episode.pubdate)
            .replace("{copyright}", episode.copyright)
        )
        subprocess.Popen(command, shell=True)

    def show_episode_url(self, episode: Episode) -> None:
        """Show episode URL in status line.

        :param episode episode to show URL for
        """
        if episode is not None:
            max_width = self._footer_window.getmaxyx()[1] - 1
            self._footer_window.addstr(1, 0, " " * max_width)
            self.change_status(episode.enclosure)

    def clear(self) -> None:
        """Clear the screen."""
        self._stdscr.clear()

    def refresh(self) -> None:
        """Refresh the screen and all windows in all perspectives."""
        self._stdscr.refresh()

        for perspective_id in self._perspectives:
            self._perspectives[perspective_id].refresh()

        self._header_window.refresh()
        self._footer_window.refresh()

    def terminate(self) -> None:
        """Set console settings to their normal state.

        This method does not, by itself, cause the application to exit. Nor
        does it even cause the input loop to end. It should simply be seen as
        a "wrapping up" method for any actions which need to be performed
        before the object is destroyed.
        """
        self._queue.stop()

        self.database.replace_queue(self._queue)
        self.database.close()

        curses.nocbreak()
        self._stdscr.keypad(False)
        curses.echo()
        curses.endwin()

    def update_parent_dimensions(self) -> None:
        """Update _parent_x and _parent_y to the size of the console."""
        current_y, current_x = self._stdscr.getmaxyx()

        if current_y != self._parent_y or current_x != self._parent_x:
            self._parent_y, self._parent_x = current_y, current_x
            self._create_windows()
            self.menus_valid = False
            self._update_timer = 0

        if self._parent_y < self.MIN_HEIGHT:
            raise DisplaySizeError("Display height is too small")
        if self._parent_x < self.MIN_WIDTH:
            raise DisplaySizeError("Display width is too small")

    def getch(self) -> int:
        """Gets an input character from the user.

        This method returns after at most INPUT_TIMEOUT ms.

        :returns int: the character entered by the user, or -1
        """
        char = self._stdscr.getch()
        return char

    def change_status(self, status) -> None:
        """Changes the status message displayed in the footer.

        :param status the status message to display
        """
        assert isinstance(status, str)

        self._status = status
        self._status_timer = self.STATUS_TIMEOUT

    def update(self) -> None:
        """Updates all actively tracked components of this object.

        Should be called by the main loop after every input or input timeout.
        """
        # have the queue check if it needs to go to the next player
        self._queue.update()

        # check the status of any downloads
        try:
            self._download_queue.update()
        except OSError as e:
            self.change_status("OSError: %s" % str(e))
            return

        # check to see if menu contents have been invalidated
        if not self.menus_valid:
            for perspective_id in self._perspectives:
                self._perspectives[perspective_id].update_menus()
            self.menus_valid = True

        # update the header text
        max_width = self._header_window.getmaxyx()[1]
        header_str = "%s " % castero.__title__
        stats_str = "[%d%%]" % self._queue.volume
        if self._queue.first is not None:
            state = self._queue.first.state
            header_str += ["Stopped", "Playing", "Paused"][state] + ": %s" % self._queue.first.title
            if self._queue.length > 1:
                header_str += " (+%d in queue)" % (self._queue.length - 1)

            # the stats section of the header contains the volume and, if
            # something is playing, time/duration of the media
            stats_str += " [%s]" % self._queue.first.time_str

        # truncate the header string to ensure there is always space for
        # the stats to be displayed
        header_str = header_str[: max_width - len(stats_str)]
        header_str += stats_str.rjust(max_width - len(header_str))
        self._header_str = header_str[:max_width]

        # update the footer text
        footer_str = "%sPress %s for help" % (
            self._status + " -- " if len(self._status) > 0 else "",
            Config["key_help"],
        )
        max_width = self._footer_window.getmaxyx()[1] - 1
        footer_str = footer_str[:max_width]
        self._footer_str = footer_str.ljust(max_width)[:max_width]

        # decrement the status timer
        if self._status_timer > 0:
            self._status_timer -= 1
            if self._status_timer <= 0:
                # status_timer should be reset during the next change_status()
                self._status = ""

        # write any episode modifications to the database
        if len(self._modified_episodes) > 0:
            for episode in self._modified_episodes:
                self.database.replace_episode(episode._feed, episode)
                if episode.progress is None:
                    self.database.delete_progress(episode, episode.progress)
                else:
                    self.database.replace_progress(episode, episode.progress)

            self.menus_valid = False
            self._modified_episodes = []

    @property
    def parent_x(self) -> int:
        """int: the width of the parent screen, in characters"""
        return self._parent_x

    @property
    def parent_y(self) -> int:
        """int: the height of the parent screen, in characters"""
        return self._parent_y

    @property
    def database(self) -> Database:
        """Database: the user's database"""
        return self._database

    @property
    def perspectives(self) -> dict:
        """dict: the loaded Perspective's with id:perspective pairs"""
        return self._perspectives

    @property
    def queue(self) -> Queue:
        """Queue: the Queue of Player's"""
        return self._queue

    @property
    def menus_valid(self) -> bool:
        """bool: whether the menu contents are valid (!need_to_be_updated)"""
        return self._menus_valid

    @menus_valid.setter
    def menus_valid(self, menus_valid) -> None:
        self._menus_valid = menus_valid

    @property
    def modified_episodes(self) -> List[Episode]:
        """List[Episode]: database episodes to save on the next update"""
        return self._modified_episodes



from lxml import etree
from typing import List

from castero.feed import Feed, FeedDownloadError, FeedStructureError, FeedParseError


class SubscriptionsError(Exception):
    """An ambiguous error while handling the document."""


class SubscriptionsLoadError(SubscriptionsError):
    """A document could not be found at the provided file, or an IO exception
    occurred when loading the file.
    """


class SubscriptionsParseError(SubscriptionsError):
    """The document could not be parsed as an XML document."""


class SubscriptionsStructureError(SubscriptionsError):
    """The file data is not a properly structured OPML document."""


class Subscriptions:
    """The user's podcast subscriptions.

    Instances of this class represent a list of podcast subscriptions, which
    the user can import from (and export to) OPML-formatted documents.
    """

    def __init__(self) -> None:
        self._tree = None
        self._feeds = []

    def load(self, path: str) -> None:
        """Load an OPML file of subscriptions.

        :param path the location of the OPML file to load

        :raises SubscriptionsParseError: unable to parse text as an XML document
        :raises SubscriptionsLoadError: an exception occurred when attempting to
          load the file
        :raises SubscriptionsStructureError: the file data is not a properly
          structured OPML document
        """
        self._tree = None
        try:
            self._tree = etree.parse(path)
        except IOError:
            raise SubscriptionsLoadError("An I/O exception occurred when attempting to load the file")
        except etree.ParseError:
            raise SubscriptionsParseError("Unable to parse text as an XML document")

    def save(self, path: str) -> None:
        """Save an OPML file of subscriptions.

        A subscriptions document must have been loaded (with .load) or created
        (with .generate) before running this method.

        :param path the location of the OPML file to create
        :raises SubscriptionsError: attempted to save before creating document
        :raises SubscriptionsLoadError: an exception occurred when attempting to
          write the file
        """
        if self._tree is not None:
            try:
                self._tree.write(path, xml_declaration=True)
            except IOError:
                raise SubscriptionsLoadError("An I/O exception occurred when attempting to save the" " file")
        else:
            raise SubscriptionsError(
                "Attempted to save an XML document that has not been loaded or" " created"
            )

    def generate(self, feeds: List[Feed]) -> None:
        """Create subscriptions document from list of feeds.

        :param feeds the list of feeds to include in the document
        """
        builder = etree.TreeBuilder()

        builder.start("opml", {"version": "2.0"})
        builder.start("head", {})
        builder.start("title", {})
        builder.data("castero feeds")
        builder.end("title")
        builder.end("head")
        builder.start("body", {})
        for feed in feeds:
            builder.start("outline", {"type": "rss", "text": str(feed), "xmlUrl": feed.key})
            builder.end("outline")
        builder.end("body")
        builder.end("opml")

        # .close returns an Element, so we need to cast to an ElementTree
        self._tree = etree.ElementTree(builder.close())

    def parse(self) -> None:
        """Parse the XML tree into a list of feeds.

        :raises SubscriptionsStructureError: the file data is not a properly
          structured OPML document
        """
        error_msg = "The file data is not a properly structured OPML document"

        if self._tree is None:
            raise SubscriptionsStructureError(error_msg)
        body = self._tree.find("body")
        if body is None:
            raise SubscriptionsStructureError(error_msg)

        feeds_container = self._find_rss_container(body)
        if feeds_container is not None:
            self._feeds = []
            for entry in feeds_container.findall("outline"):
                try:
                    feed = Feed(url=entry.attrib["xmlUrl"])
                    self._feeds.append(feed)
                    yield feed
                except FeedDownloadError as e:
                    yield (entry.attrib["xmlUrl"], e)
                except FeedStructureError as e:
                    yield (entry.attrib["xmlUrl"], e)
                except FeedParseError as e:
                    yield (entry.attrib["xmlUrl"], e)

    def _find_rss_container(self, container):
        """Find potentially-nested container for RSS feeds.

        :param container the Element to search
        :returns Element: the first 'outline' Element containing an RSS feed
        """
        outline = container.find("outline")
        if outline is None:
            return None

        if "type" in outline.attrib and outline.attrib["type"].lower() in ["rss", "link"]:
            return container
        else:
            return self._find_rss_container(outline)

    @property
    def feeds(self) -> List[Feed]:
        """List[Feed]: the loaded feeds"""
        return self._feeds



from castero import constants
from castero.config import Config
from castero.player import Player


class Queue:
    """A FIFO ordered queue of Player instances.

    This class is also the display class' main interface for accessing
    information about the current player.
    """

    MIN_VOLUME = 0
    MAX_VOLUME = 100
    MIN_SPEED = 0.5
    MAX_SPEED = 2.0

    def __init__(self, display) -> None:
        self._players = []
        self._display = display
        self._volume = int(Config["default_volume"])
        self._speed = float(Config["default_playback_speed"])
        self._resume_rewind = int(Config["resume_rewind_distance"])
        self._sanitize_volume()
        self._sanitize_speed()

    def __getitem__(self, index):
        return self._players[index]

    def __iter__(self):
        return self._players.__iter__()

    def clear(self) -> None:
        """Clears the queue.

        It is extremely likely that the caller of this method will also want to
        stop the first player prior to calling this method.
        """
        for p in self._players:
            del p
        self._players = []

    def jump(self, player) -> None:
        """Jump to the given player.

        Players skipped are removed from the queue. If the given player is not
        actually in the queue, this method safely does nothing.

        :param player the Player to jump to
        """
        if player in self._players:
            index = self._players.index(player)
            if index > 0:
                self.stop()
                self._players = self._players[index:]

    def next(self) -> None:
        """Proceed to the next player in the queue."""
        if len(self._players) > 0:
            self._players.pop(0)

    def add(self, player) -> None:
        """Adds a player to the end of the queue."""
        assert isinstance(player, Player)

        self._players.append(player)

    def play(self) -> None:
        """Plays the first player in the queue."""
        if self.first is not None:
            self._display.modified_episodes.append(self.first.episode)
            progress = self.first.episode.progress
            if progress is None or progress == 0:
                self.first.play()
            else:
                self._play_from_progress()
            self.first.set_volume(self.volume)
            self.first.set_rate(self.speed)

    def _play_from_progress(self):
        """Seek forward to progress from start of episode"""
        progress = self.first.episode.progress
        if progress is not None and progress != 0:
            resume_point = self.first.episode.progress / constants.MILLISECONDS_IN_SECOND

            # Only rewind when state was stopped
            if self.first.state == 0:
                resume_point -= self._resume_rewind

            self.first.play_from(resume_point)

    def pause(self) -> None:
        """Pauses the first player in the queue."""
        if self.first is not None:
            self.first.pause()

    def stop(self) -> None:
        """Stops the first player in the queue."""
        if self.first is not None:
            self.first.stop()

    def toggle(self) -> None:
        """Plays or pauses the first player."""
        if self.first is not None:
            if self.first.state == 1:  # playing
                self.pause()
            else:  # paused or stopped
                self.play()

    def seek(self, direction) -> None:
        """Seeks the first player in the specified direction.

        :param direction 1 to move forward, -1 to move backward
        """
        assert direction == 1 or direction == -1

        if self.first is not None:
            distance = int(Config["seek_distance_forward" if direction == 1 else "seek_distance_backward"])
            self.first.seek(direction, distance)

    def change_rate(self, direction, display=None) -> None:
        """Increase or decrease the playback speed of the first player.

        :param direction 1 to increase, -1 to decrease
        :param display (optional) the display to write status updates to
        """
        assert direction == 1 or direction == -1

        # First we change our speed value, then we set the player speed
        # to that amount. This ensures the player speed is always derived
        # from our value.
        self._speed += 0.1 * direction
        self._sanitize_speed()

        if self.first is not None:
            self.first.set_rate(self.speed)
            # Update the display status
            self._display.change_status("Playback speed set to {:0.2f}".format(self.speed))

    def change_volume(self, direction) -> None:
        """Increase or decrease volume of the current player.

        :param direction 1 to increase, -1 to decrease
        """
        assert direction == 1 or direction == -1

        # First we change our volume value, then we set the player volume
        # to that amount. This ensures the player volume is always derived
        # from our value.
        self._volume += int(Config["volume_adjust_distance"]) * direction
        self._sanitize_volume()

        if self.first is not None:
            self.first.set_volume(self._volume)

    def remove(self, player) -> int:
        """Remove a player from the queue, if it is currently in it.

        :param player the Player to remove
        :returns int: the index of the player in the queue, or -1
        """
        result = -1
        if player in self._players:
            result = self._players.index(player)
            self._players.remove(player)
        return result

    def update(self) -> None:
        """Checks the status of the current player."""
        if self.first is not None and self.first.duration is not None:
            # sanity check the player's current time
            if self.first.duration > 0:
                if (self.first.time / constants.MILLISECONDS_IN_SECOND) + 1 >= (
                    self.first.duration / constants.MILLISECONDS_IN_SECOND
                ):
                    self.first.episode.played = True
                    self.first.episode.progress = None
                    self._display.modified_episodes.append(self.first.episode)
                    self.next()
                    self.play()

    def get_episode_progress(self):
        """Get progress of the current playing episode
        :returns Tuple: episode and its progress if currently playing, else
          None is returned
        """
        if self.first is not None:
            return (self.first.episode, self.first.time)
        return (None, None)

    def _sanitize_volume(self) -> None:
        """Ensure the volume is an acceptable value (0-100 inclusive)."""
        if self._volume > self.MAX_VOLUME:
            self._volume = self.MAX_VOLUME
        elif self._volume < self.MIN_VOLUME:
            self._volume = self.MIN_VOLUME

    def _sanitize_speed(self) -> None:
        """Ensure the speed is an acceptable value (0.5-2.0 inclusive)"""
        if self._speed < self.MIN_SPEED:
            self._speed = self.MIN_SPEED
        elif self._speed > self.MAX_SPEED:
            self._speed = self.MAX_SPEED

    @property
    def first(self) -> Player:
        """Player: the first player in the queue"""
        result = None
        if len(self._players) > 0:
            result = self._players[0]
        return result

    @property
    def length(self) -> int:
        """int: the length of the queue"""
        return len(self._players)

    @property
    def volume(self) -> int:
        """int: the current playback volume"""
        return self._volume

    @property
    def speed(self) -> float:
        """float: the current playback speed"""
        return self._speed



import curses
import threading
import os

from castero.config import Config
from castero.datafile import DataFile
from castero.episode import Episode
from castero.menu import Menu


class DownloadedMenu(Menu):
    """The menu for episodes in a feed."""

    def __init__(self, window, source, child=None, active=False) -> None:
        super().__init__(window, source, child=child, active=active)

        self._episodes = []

    def __len__(self) -> int:
        return len(self._filtered_episodes)

    @property
    def _items(self):
        """A list of items in the menu represented as dictionaries."""
        result = []
        for episode in self._filtered_episodes:
            result.append(
                {
                    "attr": curses.color_pair(5) if episode.played else curses.A_NORMAL,
                    "tags": [],
                    "text": "[%s] %s" % (episode.feed_str, str(episode)),
                }
            )
        return result

    @property
    def title(self) -> str:
        """The title of the menu to display in the window header."""
        base = "Downloaded Episodes"
        if len(self._filtered_episodes) > 0:
            unplayed_episodes = 0
            for episode in self._filtered_episodes:
                if not episode.played:
                    unplayed_episodes += 1

            return "%s (%d/%d)" % (base, unplayed_episodes, len(self._filtered_episodes))
        return base

    @property
    def item(self) -> Episode:
        """The selected episode."""
        if len(self._filtered_episodes) == 0:
            return None

        return self._filtered_episodes[self._selected]

    @property
    def metadata(self) -> str:
        """Metadata for the selected episode."""
        if len(self._filtered_episodes) == 0:
            return ""

        return self._filtered_episodes[self._selected].metadata

    def update_items(self, obj):
        """Called by the parent menu (if we have one) to update our items."""
        super().update_items(obj)

        t = threading.Thread(target=self._find_downloaded_episodes, name="find_downloaded_episodes")
        t.start()

        self._sanitize()

    def _find_downloaded_episodes(self):
        if Config is None or Config["custom_download_dir"] == "":
            path = DataFile.DEFAULT_DOWNLOADED_DIR
        else:
            path = os.path.expandvars(os.path.expanduser(Config["custom_download_dir"]))
            if not path.startswith("/"):
                path = "/%s" % path

        self._episodes = []
        for (dirpath, dirnames, filenames) in os.walk(path):
            for filename in filenames:
                ep_id = filename.split("-")[0]
                if ep_id.isdigit():
                    episode = self._source.episode(int(ep_id))
                    if episode is not None:
                        self._episodes.append(episode)

        if self._inverted:
            self._episodes.reverse()
        self._sanitize()
        self.display()

    def update_child(self):
        """Not necessary for this menu -- does nothing."""
        pass

    def invert(self):
        """Invert the menu order."""
        super().invert()

        self.update_items(None)

    @property
    def _filtered_episodes(self):
        """A list of episodes which match the menu filter."""
        return list(filter(lambda ep: self._filter_text in str(ep).lower(), self._episodes))



import curses
from abc import ABC, abstractmethod

import cjkwrap

from castero.config import Config
from castero.menu import Menu


class Perspective(ABC):
    """Extendable class for display "screens".

    This class is extended by perspectives -- classes which offer methods to
    handle display elements with a certain layout. Perspectives only control
    the "inside" elements of the display: the header and footer are controlled
    by the display class.

    The user is able to switch between perspectives by using the 0-9 keys
    corresponding to the perspective's ID.

    Instances of this class do not generally hold data variables, e.g. the
    instance of the Database class. We instead reference the variables held in
    the Display instance.
    """

    ID = -1

    @abstractmethod
    def __init__(self, display):
        """
        This method does not automatically create configure some of its
        necessary elements, i.e. the windows. Instead, for some methods it is
        expected that the Display instance will call identically-named methods
        for all of its "children" perspectives when "chain" methods are called.

        For example, the Display class has a _create_windows method, which
        should call this class's create_windows method.

        :param display the parent Display instance
        """
        self._display = display

    @abstractmethod
    def create_windows(self) -> None:
        """Create and set basic parameters for the windows."""

    @abstractmethod
    def create_menus(self) -> None:
        """Create the menus used in each window.

        Windows which have menus should be created prior to running this method
        (using _create_windows).
        """

    @abstractmethod
    def display(self) -> None:
        """Draws all windows and sub-features, including titles and borders."""

    @abstractmethod
    def display_all(self) -> None:
        """Force all windows to completely redraw their content.

        The normal display() method, which is used in the core UI loop, does
        not always redraw the contents of sub-features on each iteration. For
        example, it often does not draw menu contents because they only
        typically need to be drawn when the user moves within them.

        However, if the screen is cleared, the contents of these sub-features
        must be forcefully redrawn. That is what this method provides.
        """

    @abstractmethod
    def refresh(self) -> None:
        """Refresh the screen and all windows."""

    @abstractmethod
    def handle_input(self, c) -> bool:
        """Performs action corresponding to the user's input.

        :param c the input character
        :returns bool: whether or not the application should continue running
        """

    @abstractmethod
    def made_active(self) -> None:
        """Called each time the perspective is made active (switched to)."""

    @abstractmethod
    def update_menus(self) -> None:
        """Update/refresh the contents of all menus."""

    @abstractmethod
    def _invert_selected_menu(self) -> None:
        """Inverts the contents of the selected menu."""

    @abstractmethod
    def _get_active_menu(self) -> Menu:
        """Retrieve the active Menu, if there is one.

        :returns Menu: the active Menu, or None
        """

    def _generic_handle_input(self, c) -> bool:
        """Generic handler for performing actions corresponding to input.

        :param c the input character
        :returns bool: whether or not the application should continue running
        """
        queue = self._display.queue
        key_mapping = self._display.KEY_MAPPING

        keep_running = True
        if c == key_mapping[Config["key_exit"]]:
            self.update_current_episode_progress()
            self._display.terminate()
            keep_running = False
        elif c == key_mapping[Config["key_help"]]:
            self._display.show_help()
        elif c == key_mapping[Config["key_right"]]:
            self._change_active_window(1)
            self._metadata_updated = False
        elif c == key_mapping[Config["key_left"]]:
            self._change_active_window(-1)
            self._metadata_updated = False
        elif c == key_mapping[Config["key_up"]]:
            self._get_active_menu().move(1)
            self._metadata_updated = False
        elif c == key_mapping[Config["key_down"]]:
            self._get_active_menu().move(-1)
            self._metadata_updated = False
        elif c == key_mapping[Config["key_scroll_up"]]:
            self._get_active_menu().move_page(1)
            self._metadata_updated = False
        elif c == key_mapping[Config["key_scroll_down"]]:
            self._get_active_menu().move_page(-1)
            self._metadata_updated = False
        elif c == key_mapping[Config["key_clear"]]:
            self.update_current_episode_progress()
            queue.stop()
            queue.clear()
        elif c == key_mapping[Config["key_pause_play"]] or c == key_mapping[Config["key_pause_play_alt"]]:
            self.update_current_episode_progress()
            queue.toggle()
        elif c == key_mapping[Config["key_next"]]:
            self.update_current_episode_progress()
            queue.stop()
            queue.next()
            queue.play()
        elif (
            c == key_mapping[Config["key_seek_forward"]] or c == key_mapping[Config["key_seek_forward_alt"]]
        ):
            queue.seek(1)
            self.update_current_episode_progress()
        elif (
            c == key_mapping[Config["key_seek_backward"]]
            or c == key_mapping[Config["key_seek_backward_alt"]]
        ):
            queue.seek(-1)
            self.update_current_episode_progress()
        elif c == key_mapping[Config["key_volume_increase"]]:
            queue.change_volume(1)
        elif c == key_mapping[Config["key_volume_decrease"]]:
            queue.change_volume(-1)
        elif c == key_mapping[Config["key_rate_increase"]]:
            queue.change_rate(1, display=self._display)
        elif c == key_mapping[Config["key_rate_decrease"]]:
            queue.change_rate(-1, display=self._display)
        elif c == key_mapping[Config["key_add_feed"]]:
            self._display.add_feed()
        elif c == key_mapping[Config["key_remove"]]:
            if self._active_window == 0:
                self._display.delete_feed(self._feed_menu.item)
                self.update_menus()
        elif c == key_mapping[Config["key_reload"]]:
            self._display.reload_feeds()
        elif c == key_mapping[Config["key_reload_selected"]]:
            feed = self._feed_menu.item
            if feed is not None:
                self._display.reload_selected_feed(feed)
        elif c == key_mapping[Config["key_show_url"]]:
            if self._active_window == 1 and self._episode_menu.item:
                self._display.show_episode_url(self._episode_menu.item)
        elif c == key_mapping[Config["key_save"]]:
            if self._active_window == 0 and self._feed_menu.item:
                self._display.save_episodes(feed=self._feed_menu.item)
            elif self._active_window == 1 and self._episode_menu.item:
                self._display.save_episodes(episode=self._episode_menu.item)
        elif c == key_mapping[Config["key_delete"]]:
            if self._active_window == 0 and self._feed_menu.item:
                self._display.delete_episodes(feed=self._feed_menu.item)
            elif self._active_window == 1 and self._episode_menu.item:
                self._display.delete_episodes(episode=self._episode_menu.item)
        elif c == key_mapping[Config["key_execute"]]:
            if self._active_window == 1 and self._episode_menu.item:
                self._display.execute_command(self._episode_menu.item)
        elif c == key_mapping[Config["key_invert"]]:
            self._invert_selected_menu()
        elif c == key_mapping[Config["key_filter"]]:
            menu = self._get_active_menu()
            if menu.filter_text:
                menu.filter_text = ""
            else:
                self._display.filter_menu(menu)
            self.update_menus()
            menu.move(-1)
            menu.move(1)
        elif c == key_mapping[Config["key_mark_played"]]:
            if self._active_window == 0:
                feed = self._feed_menu.item
                if feed is not None:
                    episodes = self._display.database.episodes(feed)
                    for episode in episodes:
                        episode.played = not episode.played
                        self._clear_episode_progress(episode)
                    self._display.modified_episodes.extend(episodes)
            elif self._active_window == 1:
                episode = self._episode_menu.item
                if episode is not None:
                    episode.played = not episode.played
                    self._display.modified_episodes.append(episode)
                    self._episode_menu.move(-1)
                    self._clear_episode_progress(episode)

        return keep_running

    def _change_active_window(self, direction) -> None:
        """Changes _active_window to the next or previous window, if available.

        :param direction 1 to change to the next window, -1 to change to the
          previous (if it exists)
        """
        assert direction == 1 or direction == -1

        active_menu = self._get_active_menu()
        if active_menu is not None:
            active_menu.set_active(False)

        self._active_window += direction
        if self._active_window > 1:
            self._active_window = 1
        elif self._active_window < 0:
            self._active_window = 0

        new_active_menu = self._get_active_menu()
        if new_active_menu is not None:
            new_active_menu.set_active(True)

    def _draw_metadata(self, window) -> None:
        """Draws the metadata of the selected feed/episode onto the window.

        :param window the curses window which will display the metadata
        """
        assert window is not None

        max_lines = window.getmaxyx()[0] - 4
        max_line_width = window.getmaxyx()[1] - 1

        # clear the window by drawing blank lines
        for y in range(2, window.getmaxyx()[0] - 2):
            window.addstr(y, 0, " " * max_line_width)

        metadata = self._get_active_menu().metadata
        temp_lines = metadata.split("\n")

        lines = []
        for line in temp_lines:
            parts = cjkwrap.wrap(line, window.getmaxyx()[1] - 1, replace_whitespace=True)
            if len(parts) == 0:
                lines.append("")
            else:
                for part in parts:
                    lines.append(part.strip())

        y = 2
        for line in lines[:max_lines]:
            attr = curses.color_pair(1)
            if line.startswith("!cb"):
                attr |= curses.A_BOLD
                line = line[3:]

            window.addstr(y, 0, line, attr)
            y += 1

    def update_current_episode_progress(self) -> None:
        """Update progress of the first player in queue"""
        (episode, progress) = self._display.queue.get_episode_progress()
        if episode is not None and progress is not None:
            episode.progress = progress
            self._display.database.replace_progress(episode, progress)

    def _clear_episode_progress(self, episode) -> None:
        """remove progress of the episode

        :param episode the episode to clear progress from
        """
        self._display.database.delete_progress(episode)



import curses

from castero import helpers
from castero.config import Config
from castero.menu import Menu
from castero.menus.queuemenu import QueueMenu
from castero.perspective import Perspective


class QueuePerspective(Perspective):
    """The queue-list perspective.

    This class handles display elements while in the queue perspective, which
    is a listing of the user's current Queue in which they can directly modify
    its state.
    """

    ID = 2

    def __init__(self, display) -> None:
        super().__init__(display)
        self._active_window = 0
        self._queue_window = None
        self._metadata_window = None
        self._queue_menu = None
        self._metadata_updated = False

    def create_windows(self) -> None:
        """Create and set basic parameters for the windows."""
        # delete old windows if they exist
        if self._queue_window is not None:
            del self._queue_window
            self._queue_window = None
        if self._metadata_window is not None:
            del self._metadata_window
            self._metadata_window = None

        parent_x = self._display.parent_x
        parent_y = self._display.parent_y
        third_x = helpers.third(parent_x)
        self._queue_window = curses.newwin(parent_y - 2, third_x * 2, 2, 0)
        metadata_width = parent_x - ((third_x * 2) - 1)
        self._metadata_window = curses.newwin(parent_y - 3, metadata_width, 2, 2 * third_x)

        # update menus if necessary
        if self._queue_menu is not None:
            self._queue_menu.window = self._queue_window

    def create_menus(self) -> None:
        """Create the menus used in each window."""
        assert self._queue_window is not None

        self._queue_menu = QueueMenu(self._queue_window, self._display.queue, active=True)

    def display(self) -> None:
        """Draws all windows and sub-features, including titles and borders."""
        # clear dynamic menu headers
        self._queue_window.addstr(0, 0, " " * self._queue_window.getmaxyx()[1])

        # add window headers
        self._queue_window.addstr(0, 0, self._queue_menu.title, curses.color_pair(7) | curses.A_BOLD)
        self._metadata_window.addstr(0, 0, "Metadata", curses.color_pair(7) | curses.A_BOLD)

        # add window borders
        self._queue_window.hline(
            1, 0, 0, self._queue_window.getmaxyx()[1], curses.ACS_HLINE | curses.color_pair(8)
        )
        self._metadata_window.hline(
            1, 0, 0, self._metadata_window.getmaxyx()[1] - 1, curses.ACS_HLINE | curses.color_pair(8)
        )
        if not helpers.is_true(Config["disable_vertical_borders"]):
            self._queue_window.vline(
                0,
                self._queue_window.getmaxyx()[1] - 1,
                0,
                self._queue_window.getmaxyx()[0] - 2,
                curses.ACS_VLINE | curses.color_pair(8),
            )

        # draw metadata
        if not self._metadata_updated:
            self._draw_metadata(self._metadata_window)
            self._metadata_window.refresh()
            self._metadata_updated = True

        self._queue_window.refresh()

    def display_all(self) -> None:
        """Force all windows to completely redraw their content."""
        self._metadata_updated = False
        self._queue_menu.display()
        self.display()

    def handle_input(self, c) -> bool:
        """Performs action corresponding to the user's input."""
        queue = self._display.queue
        key_mapping = self._display.KEY_MAPPING

        keep_running = True
        if c == key_mapping[Config["key_play_selected"]]:
            target = self._queue_menu.item
            self._display.queue.jump(target)
            while self._queue_menu.selected_index > 0:
                self._queue_menu.move(1)
            queue.play()
            self._display.menus_valid = False
        elif c == key_mapping[Config["key_next"]]:
            queue.stop()
            queue.next()
            self._queue_menu.move(1)
            queue.play()
            self._display.menus_valid = False
        elif c == key_mapping[Config["key_clear"]]:
            queue.stop()
            queue.clear()
            self._display.menus_valid = False
        elif c == key_mapping[Config["key_remove"]]:
            self._remove_selected_from_queue()
        elif c == key_mapping[Config["key_show_url"]]:
            item = self._queue_menu.item
            if item is not None:
                self._display.show_episode_url(item.episode)
        elif c == key_mapping[Config["key_execute"]]:
            item = self._queue_menu.item
            if item is not None:
                self._display.execute_command(item.episode)
        elif c == key_mapping[Config["key_reload_selected"]]:
            pass
        elif c == key_mapping[Config["key_save"]]:
            pass
        elif c == key_mapping[Config["key_delete"]]:
            pass
        elif c == key_mapping[Config["key_mark_played"]]:
            pass
        elif c == key_mapping[Config["key_filter"]]:
            pass
        else:
            keep_running = self._generic_handle_input(c)

        return keep_running

    def refresh(self) -> None:
        """Refresh the screen and all windows."""
        self._queue_window.refresh()
        self._metadata_window.refresh()
        self._queue_menu.refresh()

    def made_active(self) -> None:
        """Called each time the perspective is made active (switched to)."""
        pass

    def update_menus(self) -> None:
        """Update/refresh the contents of all menus."""
        self._queue_menu.update_items(None)
        self._metadata_updated = False

    def _get_active_menu(self) -> Menu:
        """Retrieve the active Menu, if there is one."""
        assert 0 <= self._active_window < 2

        return self._queue_menu

    def _invert_selected_menu(self) -> None:
        """Inverts the contents of the selected menu."""
        pass

    def _remove_selected_from_queue(self) -> None:
        """Remove the selected player from the queue."""
        player = self._queue_menu.item
        if player is not None:
            index = self._display.queue.remove(player)
            self._queue_menu.update_items(None)
            for i in range(index):
                self._get_active_menu().move(-1)
        self._metadata_updated = False



__title__ = "castero"
__description__ = "A TUI podcast client for the terminal"
__keywords__ = "podcast commandline terminal tui curses"
__author__ = "xgi"
__author_email__ = "xgi@faltro.com"
__version__ = "0.9.5"
__date__ = "2022-04-02"
__copyright__ = "Copyright (c) 2018 Jake Robertson"
__license__ = "MIT License"
__url__ = "https://github.com/xgi/castero"

__help__ = """\
%s help
============

Version: %s
Updated: %s
Maintainer: %s <%s>
License: %s
URL: %s

Commands
    {key_help} - show this help screen
    {key_exit} - exit the client
    {key_add_feed} - add a feed
    {key_remove} - remove the selected feed
    {key_reload} - reload all feeds
    {key_reload_selected} - reload the selected feed
    {key_save} - save episode for offline playback
    {key_delete} - delete downloaded episodes
    {key_up/key_down} - navigate up/down in menus
    {key_right/key_left} - navigate right/left in menus
    {key_scroll_up/key_scroll_down} - scroll up/down in menus
    {key_play_selected} - play selected feed/episode
    {key_add_selected} - add selected feed/episode to queue
    {key_clear} - clear the queue
    {key_clear_progress} - clear progress of selected episode
    {key_next} - go to the next episode in the queue
    {key_invert} - invert the order of the menu
    {key_filter} - filter the contents of the menu
    {key_mark_played} - mark episode as played/unplayed
    {key_pause_play|key_pause_play_alt} - pause/play the current episode
    {key_seek_forward|key_seek_forward_alt} - seek forward
    {key_seek_backward|key_seek_backward_alt} - seek backward
    {key_volume_increase/key_volume_decrease} - increase/decrease volume
    {key_rate_increase/key_rate_decrease} - increase/decrease playback speed
    {key_show_url} - show episode URL
    {1-5} - change between client layouts
""" % (
    __title__,
    __version__,
    __date__,
    __author__,
    __author_email__,
    __license__,
    __url__,
)



import curses

from castero import helpers
from castero.config import Config
from castero.menu import Menu
from castero.menus.episodemenu import EpisodeMenu
from castero.menus.feedmenu import FeedMenu
from castero.perspective import Perspective
from castero.player import Player


class SimplePerspective(Perspective):
    """The simple perspective, similar to the primary one.

    This class handles display elements while in the simple perspective, which
    is similar to the primary perspective but without a metadata window.
    """

    ID = 3

    def __init__(self, display) -> None:
        super().__init__(display)
        self._active_window = 0
        self._feed_window = None
        self._episode_window = None
        self._feed_menu = None
        self._episode_menu = None
        self._queue_unplayed_feed_episodes = helpers.is_true(Config["add_only_unplayed_episodes"])

    def create_windows(self) -> None:
        """Create and set basic parameters for the windows."""
        # delete old windows if they exist
        if self._feed_window is not None:
            del self._feed_window
            self._feed_window = None
        if self._episode_window is not None:
            del self._episode_window
            self._episode_window = None

        parent_x = self._display.parent_x
        parent_y = self._display.parent_y

        fw_width = min(25, parent_x // 2)

        self._feed_window = curses.newwin(parent_y - 2, fw_width, 2, 0)
        self._episode_window = curses.newwin(parent_y - 2, parent_x - fw_width, 2, fw_width)

        # update menus if necessary
        if self._feed_menu is not None:
            self._feed_menu.window = self._feed_window
        if self._episode_menu is not None:
            self._episode_menu.window = self._episode_window

    def create_menus(self) -> None:
        """Create the menus used in each window, if necessary."""
        assert all(window is not None for window in [self._feed_window, self._episode_window])

        self._episode_menu = EpisodeMenu(self._episode_window, self._display.database)
        self._feed_menu = FeedMenu(
            self._feed_window, self._display.database, child=self._episode_menu, active=True
        )

    def display(self) -> None:
        """Draws all windows and sub-features, including titles and borders."""
        # clear dynamic menu headers
        self._feed_window.addstr(0, 0, " " * self._feed_window.getmaxyx()[1])
        self._episode_window.addstr(0, 0, " " * self._episode_window.getmaxyx()[1])

        # add window headers
        self._feed_window.addstr(0, 0, self._feed_menu.title, curses.color_pair(7) | curses.A_BOLD)
        self._episode_window.addstr(0, 0, self._episode_menu.title, curses.color_pair(7) | curses.A_BOLD)

        # add window borders
        self._feed_window.hline(
            1, 0, 0, self._feed_window.getmaxyx()[1], curses.ACS_HLINE | curses.color_pair(8)
        )
        self._episode_window.hline(
            1, 0, 0, self._episode_window.getmaxyx()[1], curses.ACS_HLINE | curses.color_pair(8)
        )
        if not helpers.is_true(Config["disable_vertical_borders"]):
            self._feed_window.vline(
                0,
                self._feed_window.getmaxyx()[1] - 1,
                0,
                self._feed_window.getmaxyx()[0] - 2,
                curses.ACS_VLINE | curses.color_pair(8),
            )

        self._feed_window.refresh()
        self._episode_window.refresh()

    def display_all(self) -> None:
        """Force all windows to completely redraw their content."""
        self._metadata_updated = False
        self._feed_menu.display()
        self._episode_menu.display()
        self.display()

    def handle_input(self, c) -> bool:
        """Performs action corresponding to the user's input."""
        queue = self._display.queue
        key_mapping = self._display.KEY_MAPPING

        keep_running = True
        if c == key_mapping[Config["key_play_selected"]]:
            queue.stop()
            queue.clear()
            self._create_player_from_selected()
            queue.play()
        elif c == key_mapping[Config["key_add_selected"]]:
            self._create_player_from_selected()
            self._get_active_menu().move(-1)
        elif c == key_mapping[Config["key_clear_progress"]]:
            self._clear_progress_from_selected()
        else:
            keep_running = self._generic_handle_input(c)

        return keep_running

    def made_active(self) -> None:
        """Called each time the perspective is made active (switched to)."""

    def update_menus(self) -> None:
        """Update/refresh the contents of all menus."""
        self._feed_menu.update_items(None)
        self._feed_menu.update_child()
        self._metadata_updated = False

    def refresh(self) -> None:
        """Refresh the screen and all windows."""
        self._feed_window.refresh()
        self._episode_window.refresh()
        self._feed_menu.refresh()

    def _get_active_menu(self) -> Menu:
        """Retrieve the active Menu, if there is one."""
        assert 0 <= self._active_window < 2

        return {
            0: self._feed_menu,
            1: self._episode_menu,
        }.get(self._active_window)

    def _invert_selected_menu(self) -> None:
        """Inverts the contents of the selected menu."""
        self._get_active_menu().invert()
        if self._feed_menu:
            self._feed_menu.update_child()
        self._metadata_updated = False

    def _create_player_from_selected(self) -> None:
        """Creates player(s) based on the selected items and adds to the queue.

        If the active menu is the feed menu, then this will create players for
        all episodes in the selected feed. If the active menu is the episode
        menu, this will simply create a single player.

        This method will not clear the queue prior to adding the new player(s),
        nor will it play the episodes after running.
        """
        if self._active_window == 0:
            feed = self._feed_menu.item
            if feed is not None:
                if self._queue_unplayed_feed_episodes:
                    episodes = self._display.database.unplayed_episodes(feed)
                else:
                    episodes = self._display.database.episodes(feed)

                for episode in episodes:
                    player = Player.create_instance(
                        self._display.AVAILABLE_PLAYERS, str(episode), episode.get_playable(), episode
                    )
                    self._display.queue.add(player)
        elif self._active_window == 1:
            episode = self._episode_menu.item
            if episode is not None:
                player = Player.create_instance(
                    self._display.AVAILABLE_PLAYERS, str(episode), episode.get_playable(), episode
                )
                self._display.queue.add(player)
        self._metadata_updated = False

    def _clear_progress_from_selected(self) -> None:
        if self._active_window == 1:
            episode = self._episode_menu.item
            if episode is not None:
                self._display.database.delete_progress(episode)
                self._episode_window.refresh()



import curses

from castero.player import Player
from castero.menu import Menu


class QueueMenu(Menu):
    """The menu for the player queue."""

    def __init__(self, window, source, child=None, active=False) -> None:
        super().__init__(window, source, child=child, active=active)

    def __len__(self) -> int:
        return self._source.length

    @property
    def _items(self):
        """A list of items in the menu represented as dictionaries."""
        return [{"attr": curses.A_NORMAL, "tags": [], "text": str(player)} for player in self._source]

    @property
    def title(self) -> str:
        """The title of the menu to display in the window header."""
        return "Queue"

    @property
    def item(self) -> Player:
        """The selected player."""
        if self._source.length == 0:
            return None

        return self._source[self._selected]

    @property
    def metadata(self) -> str:
        """Metadata for the selected player's episode."""
        player = self.item
        if player is None:
            return ""

        return player.episode.metadata

    def update_items(self, obj):
        """Called by the parent menu (if we have one) to update our items."""
        super().update_items(obj)
        self.display()

    def update_child(self):
        """Not necessary for this menu -- does nothing."""
        pass

    def invert(self):
        """Invert the menu order."""
        pass



import curses

from castero import helpers
from castero.config import Config
from castero.menu import Menu
from castero.menus.chronomenu import ChronoMenu
from castero.perspective import Perspective
from castero.player import Player


class ChronoPerspective(Perspective):
    """The chronological perspective.

    This class handles display elements while in the chronological perspective,
    which is a listing of episodes from all feeds ordered chronologically.
    """

    ID = 5

    def __init__(self, display) -> None:
        super().__init__(display)
        self._active_window = 0
        self._episode_window = None
        self._metadata_window = None
        self._episode_menu = None
        self._metadata_updated = False

    def create_windows(self) -> None:
        """Create and set basic parameters for the windows."""
        # delete old windows if they exist
        if self._episode_window is not None:
            del self._episode_window
            self._episode_window = None
        if self._metadata_window is not None:
            del self._metadata_window
            self._metadata_window = None

        parent_x = self._display.parent_x
        parent_y = self._display.parent_y
        third_x = helpers.third(parent_x)
        self._episode_window = curses.newwin(parent_y - 2, third_x * 2, 2, 0)
        metadata_width = parent_x - ((third_x * 2) - 1)
        self._metadata_window = curses.newwin(parent_y - 3, metadata_width, 2, 2 * third_x)

        # update menus if necessary
        if self._episode_menu is not None:
            self._episode_menu.window = self._episode_window

    def create_menus(self) -> None:
        """Create the menus used in each window."""
        assert all(window is not None for window in [self._episode_window])

        self._episode_menu = ChronoMenu(self._episode_window, self._display.database, active=True)

    def display(self) -> None:
        """Draws all windows and sub-features, including titles and borders."""
        # clear dynamic menu headers
        self._episode_window.addstr(0, 0, " " * self._episode_window.getmaxyx()[1])

        # add window headers
        self._episode_window.addstr(0, 0, self._episode_menu.title, curses.color_pair(7) | curses.A_BOLD)
        self._metadata_window.addstr(0, 0, "Metadata", curses.color_pair(7) | curses.A_BOLD)

        # add window borders
        self._episode_window.hline(
            1, 0, 0, self._episode_window.getmaxyx()[1], curses.ACS_HLINE | curses.color_pair(8)
        )
        self._metadata_window.hline(
            1, 0, 0, self._metadata_window.getmaxyx()[1] - 1, curses.ACS_HLINE | curses.color_pair(8)
        )
        if not helpers.is_true(Config["disable_vertical_borders"]):
            self._episode_window.vline(
                0,
                self._episode_window.getmaxyx()[1] - 1,
                0,
                self._episode_window.getmaxyx()[0] - 2,
                curses.ACS_VLINE | curses.color_pair(8),
            )

        # draw metadata
        if not self._metadata_updated:
            self._draw_metadata(self._metadata_window)
            self._metadata_window.refresh()
            self._metadata_updated = True

        self._episode_window.refresh()

    def display_all(self) -> None:
        """Force all windows to completely redraw their content."""
        self._metadata_updated = False
        self._episode_menu.display()
        self.display()

    def handle_input(self, c) -> bool:
        """Performs action corresponding to the user's input."""
        queue = self._display.queue
        key_mapping = self._display.KEY_MAPPING

        keep_running = True
        if c == key_mapping[Config["key_play_selected"]]:
            queue.stop()
            queue.clear()
            self._create_player_from_selected()
            queue.play()
        elif c == key_mapping[Config["key_add_selected"]]:
            self._create_player_from_selected()
            self._get_active_menu().move(-1)
        elif c == key_mapping[Config["key_show_url"]]:
            if self._episode_menu.item:
                self._display.show_episode_url(self._episode_menu.item)
        elif c == key_mapping[Config["key_save"]]:
            if self._episode_menu.item:
                self._display.save_episodes(episode=self._episode_menu.item)
                self._display.menus_valid = False
        elif c == key_mapping[Config["key_delete"]]:
            if self._episode_menu.item:
                self._display.delete_episodes(episode=self._episode_menu.item)
                self._display.menus_valid = False
        elif c == key_mapping[Config["key_mark_played"]]:
            if self._active_window == 0:
                episode = self._episode_menu.item
                if episode is not None:
                    episode.played = not episode.played
                    self._display.modified_episodes.append(episode)
                    self._episode_menu.move(-1)
        elif c == key_mapping[Config["key_execute"]]:
            episode = self._episode_menu.item
            if episode is not None:
                self._display.execute_command(episode)
        elif c == key_mapping[Config["key_reload_selected"]]:
            pass
        elif c == key_mapping[Config["key_remove"]]:
            pass
        else:
            keep_running = self._generic_handle_input(c)

        return keep_running

    def made_active(self) -> None:
        """Called each time the perspective is made active (switched to)."""

    def update_menus(self) -> None:
        """Update/refresh the contents of all menus."""
        self._episode_menu.update_items(None)
        self._metadata_updated = False

    def refresh(self) -> None:
        """Refresh the screen and all windows."""
        self._episode_window.refresh()
        self._metadata_window.refresh()
        self._episode_menu.refresh()

    def _get_active_menu(self) -> Menu:
        """Retrieve the active Menu, if there is one."""
        assert 0 <= self._active_window < 2

        return self._episode_menu

    def _invert_selected_menu(self) -> None:
        """Inverts the contents of the selected menu."""
        self._get_active_menu().invert()
        self._metadata_updated = False

    def _create_player_from_selected(self) -> None:
        """Creates player(s) based on the selected items and adds to the queue.

        If the active menu is the feed menu, then this will create players for
        all episodes in the selected feed. If the active menu is the episode
        menu, this will simply create a single player.

        This method will not clear the queue prior to adding the new player(s),
        nor will it play the episodes after running.
        """
        episode = self._episode_menu.item
        if episode is not None:
            player = Player.create_instance(
                self._display.AVAILABLE_PLAYERS, str(episode), episode.get_playable(), episode
            )
            self._display.queue.add(player)
        self._metadata_updated = False



import curses

from castero.feed import Feed
from castero.menu import Menu
from castero.menus.episodemenu import EpisodeMenu


class FeedMenu(Menu):
    """The menu for a podcast feed."""

    def __init__(self, window, source, child=None, active=False) -> None:
        assert child is not None and isinstance(child, EpisodeMenu)

        self._feeds = []

        super().__init__(window, source, child=child, active=active)

    def __len__(self) -> int:
        return len(self._filtered_feeds)

    @property
    def _items(self):
        """A list of items in the menu represented as dictionaries."""
        return [{"attr": curses.A_NORMAL, "tags": [], "text": str(feed)} for feed in self._feeds]

    @property
    def title(self) -> str:
        """The title of the menu to display in the window header."""
        return "Feeds"

    @property
    def item(self) -> Feed:
        """The selected feed."""
        if len(self._filtered_feeds) == 0:
            return None

        return self._filtered_feeds[self._selected]

    @property
    def metadata(self) -> str:
        """Metadata for the selected feed."""
        feed = self.item
        if feed is None:
            return ""

        return feed.metadata

    def update_items(self, obj):
        """Called by the parent menu (if we have one) to update our items."""
        super().update_items(obj)

        self._feeds = self._source.feeds()
        if self._inverted:
            self._feeds.reverse()

        self._sanitize()
        self.display()

    def update_child(self):
        """Update our child menu, the episode menu."""
        if len(self._filtered_feeds) == 0:
            self.update_items(None)
            self._child.update_items(None)
        else:
            self._child.update_items(self._filtered_feeds[self._selected])

    def invert(self):
        """Invert the menu order."""
        super().invert()

        self.update_items(None)

    @property
    def _filtered_feeds(self):
        """A list of feeds which match the menu filter."""
        return list(filter(lambda feed: self._filter_text in str(feed).lower(), self._feeds))
